import dataclasses
import json
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from pandas.api.types import is_string_dtype

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import DataQualityGetPlotData
from evidently.calculations.data_quality import FeatureQualityStats
from evidently.calculations.data_quality import get_features_stats
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import ColumnType
from evidently.utils.data_preprocessing import DataDefinition
from evidently.utils.types import Numeric
from evidently.utils.visualizations import plot_boxes
from evidently.utils.visualizations import plot_cat_cat_rel
from evidently.utils.visualizations import plot_cat_feature_in_time
from evidently.utils.visualizations import plot_distr
from evidently.utils.visualizations import plot_distr_with_log_button
from evidently.utils.visualizations import plot_num_feature_in_time
from evidently.utils.visualizations import plot_num_num_rel
from evidently.utils.visualizations import plot_time_feature_distr


@dataclasses.dataclass
class NumericCharacteristics:
    number_of_rows: int
    count: int
    mean: Optional[Numeric]
    std: Optional[Numeric]
    min: Optional[Numeric]
    p25: Optional[Numeric]
    p50: Optional[Numeric]
    p75: Optional[Numeric]
    max: Optional[Numeric]
    unique: Optional[int]
    unique_percentage: Optional[float]
    missing: Optional[int]
    missing_percentage: Optional[float]
    infinite_count: Optional[int]
    infinite_percentage: Optional[float]
    most_common: Optional[Union[int, float]]
    most_common_percentage: Optional[float]


@dataclasses.dataclass
class CategoricalCharacteristics:
    number_of_rows: int
    count: int
    unique: Optional[int]
    unique_percentage: Optional[float]
    most_common: Optional[object]
    most_common_percentage: Optional[float]
    missing: Optional[int]
    missing_percentage: Optional[float]
    new_in_current_values_count: Optional[int] = None
    unused_in_current_values_count: Optional[int] = None


@dataclasses.dataclass
class DatetimeCharacteristics:
    number_of_rows: int
    count: int
    unique: Optional[int]
    unique_percentage: Optional[float]
    most_common: Optional[object]
    most_common_percentage: Optional[float]
    missing: Optional[int]
    missing_percentage: Optional[float]
    first: Optional[str]
    last: Optional[str]


@dataclasses.dataclass
class TextCharacteristics:
    number_of_rows: int
    count: int
    missing: Optional[int]
    missing_percentage: Optional[float]
    text_length_min: Optional[float]
    text_length_mean: Optional[float]
    text_length_max: Optional[float]
    oov_min: Optional[float]
    oov_mean: Optional[float]
    oov_max: Optional[float]
    non_letter_char_min: Optional[float]
    non_letter_char_mean: Optional[float]
    non_letter_char_max: Optional[float]


ColumnCharacteristics = Union[
    NumericCharacteristics, CategoricalCharacteristics, DatetimeCharacteristics, TextCharacteristics
]


@dataclasses.dataclass
class DataInTime:
    data_for_plots: Dict[str, pd.DataFrame]
    freq: str
    datetime_name: str


@dataclasses.dataclass
class DataByTarget:
    data_for_plots: Dict[str, Dict[str, Union[list, pd.DataFrame]]]
    target_name: str
    target_type: str


@dataclasses.dataclass
class DataQualityPlot:
    bins_for_hist: Optional[Dict[str, pd.DataFrame]]
    data_in_time: Optional[DataInTime]
    data_by_target: Optional[DataByTarget]
    counts_of_values: Optional[Dict[str, pd.DataFrame]]


@dataclasses.dataclass
class ColumnSummary:
    column_name: str
    column_type: str
    reference_characteristics: Optional[ColumnCharacteristics]
    current_characteristics: ColumnCharacteristics
    plot_data: DataQualityPlot


class ColumnSummaryMetric(Metric[ColumnSummary]):
    column_name: str
    generated_text_features: Optional[Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]]

    def __init__(self, column_name: str):
        self.column_name = column_name
        self.generated_text_features = None

    def required_features(self, data_definition: DataDefinition):
        column_type = data_definition.get_column(self.column_name).column_type
        if column_type == ColumnType.Text:
            self.generated_text_features = {}
            self.generated_text_features["text_length"] = TextLength(self.column_name)
            self.generated_text_features["non_letter_char"] = NonLetterCharacterPercentage(self.column_name)
            self.generated_text_features["oov"] = OOVWordsPercentage(self.column_name)
            return list(self.generated_text_features.values())
        return []

    def get_parameters(self) -> tuple:
        return (self.column_name,)

    @staticmethod
    def acceptable_types() -> List[ColumnType]:
        return [ColumnType.Numerical, ColumnType.Categorical, ColumnType.Text]

    def calculate(self, data: InputData) -> ColumnSummary:
        columns = process_columns(data.current_data, data.column_mapping)

        if self.column_name not in data.current_data:
            raise ValueError(f"Column '{self.column_name}' not found in current dataset.")

        if data.reference_data is not None and self.column_name not in data.reference_data:
            raise ValueError(f"Column '{self.column_name}' not found in reference dataset.")

        column_type = None
        target_name = columns.utility_columns.target
        target_type = None
        data_by_target = None

        # define target type and prediction type. TODO move it to process_columns func
        if columns.utility_columns.target is not None:
            reg_condition = data.column_mapping.task == "regression" or (
                is_numeric_dtype(data.current_data[target_name])
                and columns.task != "classification"
                and data.current_data[target_name].nunique() > 5
            )
            if reg_condition:
                target_type = "num"
            else:
                target_type = "cat"
            if target_name == self.column_name:
                column_type = target_type

        if columns.utility_columns.prediction is not None:
            if (
                isinstance(columns.utility_columns.prediction, str)
                and columns.utility_columns.prediction == self.column_name
            ):
                if (
                    is_string_dtype(data.current_data[columns.utility_columns.prediction])
                    or (
                        is_numeric_dtype(data.current_data[columns.utility_columns.prediction])
                        and columns.task != "classification"
                        and data.current_data[columns.utility_columns.prediction].nunique() < 5
                    )
                    or (
                        is_numeric_dtype(data.current_data[columns.utility_columns.prediction])
                        and columns.task == "classification"
                        and (
                            data.current_data[columns.utility_columns.prediction].max() > 1
                            or data.current_data[columns.utility_columns.prediction].min() < 0
                        )
                    )
                ):
                    column_type = "cat"
                else:
                    column_type = "num"

            if (
                isinstance(columns.utility_columns.prediction, list)
                and self.column_name in columns.utility_columns.prediction
            ):
                column_type = "num"
        if self.column_name in columns.num_feature_names:
            column_type = "num"
        elif self.column_name in columns.cat_feature_names:
            column_type = "cat"
        elif self.column_name in columns.datetime_feature_names or (
            columns.utility_columns.date is not None and columns.utility_columns.date == self.column_name
        ):
            column_type = "datetime"
        elif self.column_name in columns.text_feature_names:
            column_type = "text"
        if column_type is None:
            raise ValueError(f"column {self.column_name} not in num, cat, text or datetime features lists")

        curr_characteristics: ColumnCharacteristics
        reference_data = None
        ref_characteristics: Optional[ColumnCharacteristics] = None
        if column_type == "text" and self.generated_text_features is not None:
            if data.reference_data is not None:
                ref_characteristics = self.get_text_stats(
                    "reference",
                    data,
                    data.reference_data[self.column_name],
                    self.generated_text_features,
                )
            curr_characteristics = self.get_text_stats(
                "current",
                data,
                data.current_data[self.column_name],
                self.generated_text_features,
            )
        else:
            if data.reference_data is not None:
                reference_data = data.reference_data
                ref_characteristics = self.map_data(
                    get_features_stats(data.reference_data[self.column_name], column_type)
                )
            curr_characteristics = self.map_data(get_features_stats(data.current_data[self.column_name], column_type))

            if data.reference_data is not None and column_type == "cat":
                current_values_set = set(data.current_data[self.column_name].unique())
                reference_values_set = set(data.reference_data[self.column_name].unique())
                unique_in_current = current_values_set - reference_values_set
                new_in_current_values_count: int = len(unique_in_current)
                unique_in_reference = reference_values_set - current_values_set
                unused_in_current_values_count: int = len(unique_in_reference)
                if any(pd.isnull(list(unique_in_current))) and any(pd.isnull(list(unique_in_reference))):
                    new_in_current_values_count -= 1
                    unused_in_current_values_count -= 1
                if not isinstance(curr_characteristics, CategoricalCharacteristics):
                    raise ValueError(f"{self.column_name} should be categorical")
                curr_characteristics.new_in_current_values_count = new_in_current_values_count
                curr_characteristics.unused_in_current_values_count = unused_in_current_values_count

        # plot data
        gpd = DataQualityGetPlotData()
        bins_for_hist = gpd.calculate_main_plot(data.current_data, reference_data, self.column_name, column_type)
        data_in_time = None
        if (
            columns.utility_columns.date is not None
            and columns.utility_columns.date != self.column_name
            and column_type != "datetime"
        ):
            data_in_time = gpd.calculate_data_in_time(
                data.current_data,
                reference_data,
                self.column_name,
                column_type,
                columns.utility_columns.date,
            )
            data_in_time = DataInTime(
                data_for_plots={
                    "current": data_in_time["current"],
                    "reference": data_in_time["reference"],
                },
                freq=data_in_time["freq"],
                datetime_name=data_in_time["datetime_name"],
            )

        if (
            target_name is not None
            and target_type is not None
            and columns.utility_columns.target != self.column_name
            and column_type != "datetime"
        ):

            data_for_plots = gpd.calculate_data_by_target(
                data.current_data,
                reference_data,
                self.column_name,
                column_type,
                target_name,
                target_type,
            )
            data_by_target = DataByTarget(
                data_for_plots=data_for_plots,
                target_name=target_name,
                target_type=target_type,
            )
        counts_of_values = None
        if column_type in ["cat", "num"]:
            counts_of_values = {}
            current_counts = data.current_data[self.column_name].value_counts(dropna=False).reset_index()
            current_counts.columns = ["x", "count"]
            counts_of_values["current"] = current_counts.head(10)
            if reference_data is not None:
                reference_counts = data.reference_data[self.column_name].value_counts(dropna=False).reset_index()
                reference_counts.columns = ["x", "count"]
                counts_of_values["reference"] = reference_counts.head(10)

        plot_data = DataQualityPlot(
            bins_for_hist=bins_for_hist,
            data_in_time=data_in_time,
            data_by_target=data_by_target,
            counts_of_values=counts_of_values,
        )

        return ColumnSummary(
            column_name=self.column_name,
            column_type=column_type,
            reference_characteristics=ref_characteristics,
            current_characteristics=curr_characteristics,
            plot_data=plot_data,
        )

    @staticmethod
    def map_data(stats: FeatureQualityStats) -> ColumnCharacteristics:
        if stats.feature_type == "num":
            if isinstance(stats.max, str) or isinstance(stats.min, str) or isinstance(stats.most_common_value, str):
                raise ValueError("max / min stats should be int or float type, but got str")
            return NumericCharacteristics(
                number_of_rows=stats.number_of_rows,
                count=stats.count,
                mean=stats.mean,
                std=stats.std,
                min=stats.min,
                max=stats.max,
                p25=stats.percentile_25,
                p50=stats.percentile_50,
                p75=stats.percentile_75,
                unique=stats.unique_count,
                unique_percentage=stats.unique_percentage,
                missing=stats.missing_count,
                missing_percentage=stats.missing_percentage,
                infinite_count=stats.infinite_count,
                infinite_percentage=stats.infinite_percentage,
                most_common=stats.most_common_value,
                most_common_percentage=stats.most_common_value_percentage,
            )
        if stats.feature_type == "cat":
            return CategoricalCharacteristics(
                number_of_rows=stats.number_of_rows,
                count=stats.count,
                unique=stats.unique_count,
                unique_percentage=stats.unique_percentage,
                most_common=stats.most_common_value,
                most_common_percentage=stats.most_common_value_percentage,
                missing=stats.missing_count,
                missing_percentage=stats.missing_percentage,
            )
        if stats.feature_type == "datetime":
            if not isinstance(stats.min, str) or not isinstance(stats.max, str):
                raise ValueError(f"min / max expected to be str for datetime, got {type(stats.min)}/{type(stats.max)}")
            return DatetimeCharacteristics(
                number_of_rows=stats.number_of_rows,
                count=stats.count,
                unique=stats.unique_count,
                unique_percentage=stats.unique_percentage,
                most_common=stats.most_common_value,
                most_common_percentage=stats.most_common_value_percentage,
                missing=stats.missing_count,
                missing_percentage=stats.missing_percentage,
                first=stats.min,
                last=stats.max,
            )
        raise ValueError(f"unknown feature type {stats.feature_type}")

    def get_text_stats(
        self,
        dataset: str,
        data: InputData,
        text_feature: pd.Series,
        generated_text_features: dict,
    ) -> TextCharacteristics:
        number_of_rows = len(text_feature)
        missing = text_feature.isna().sum()
        if dataset == "current":
            text_length = data.get_current_column(generated_text_features["text_length"].feature_name())
            oov = data.get_current_column(generated_text_features["oov"].feature_name())
            non_letter_char = data.get_current_column(generated_text_features["non_letter_char"].feature_name())
        else:
            text_length = data.get_reference_column(generated_text_features["text_length"].feature_name())
            oov = data.get_reference_column(generated_text_features["oov"].feature_name())
            non_letter_char = data.get_reference_column(generated_text_features["non_letter_char"].feature_name())

        return TextCharacteristics(
            number_of_rows=number_of_rows,
            count=text_feature.count(),
            missing=missing,
            missing_percentage=np.round(missing / number_of_rows, 3),
            text_length_min=text_length.min(),
            text_length_mean=np.round(text_length.mean(), 3),
            text_length_max=text_length.max(),
            oov_min=np.round(oov.min(), 3),
            oov_mean=np.round(oov.mean(), 3),
            oov_max=np.round(oov.max(), 3),
            non_letter_char_min=np.round(non_letter_char.min(), 3),
            non_letter_char_mean=np.round(non_letter_char.mean(), 3),
            non_letter_char_max=np.round(non_letter_char.max(), 3),
        )


@default_renderer(wrap_type=ColumnSummaryMetric)
class ColumnSummaryMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnSummaryMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        result.pop("plot_data", None)
        return result

    def render_html(self, obj: ColumnSummaryMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        column_type = metric_result.column_type
        column_name = metric_result.column_name
        # main plot
        bins_for_hist = metric_result.plot_data.bins_for_hist
        if bins_for_hist is not None:
            hist_curr = bins_for_hist["current"]
            hist_ref = None
            metrics_values_headers = [""]
            if "reference" in bins_for_hist.keys():
                hist_ref = bins_for_hist["reference"]
                metrics_values_headers = ["current", "reference"]

            if column_type == "cat":
                fig = plot_distr(hist_curr=hist_curr, hist_ref=hist_ref, color_options=self.color_options)
                fig.update_layout(legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1))
                fig = json.loads(fig.to_json())
            if column_type == "num":
                ref_log = None
                if "reference_log" in bins_for_hist.keys():
                    ref_log = bins_for_hist["reference_log"]
                fig = plot_distr_with_log_button(
                    hist_curr, bins_for_hist["current_log"], hist_ref, ref_log, color_options=self.color_options
                )
            if column_type == "datetime":
                fig = plot_time_feature_distr(hist_curr, hist_ref, column_name, color_options=self.color_options)
            graph = {"data": fig["data"], "layout": fig["layout"]}
        else:
            graph = {}
            metrics_values_headers = [""]
            if metric_result.reference_characteristics is not None:
                metrics_values_headers = ["current", "reference"]

        # additional plots
        additional_graphs = []
        parts = []
        if metric_result.plot_data.data_in_time is not None:
            if column_type == "num":
                feature_in_time_figure = plot_num_feature_in_time(
                    metric_result.plot_data.data_in_time.data_for_plots["current"],
                    metric_result.plot_data.data_in_time.data_for_plots["reference"],
                    column_name,
                    metric_result.plot_data.data_in_time.datetime_name,
                    metric_result.plot_data.data_in_time.freq,
                    color_options=self.color_options,
                )
            if column_type == "cat":
                feature_in_time_figure = plot_cat_feature_in_time(
                    metric_result.plot_data.data_in_time.data_for_plots["current"],
                    metric_result.plot_data.data_in_time.data_for_plots["reference"],
                    column_name,
                    metric_result.plot_data.data_in_time.datetime_name,
                    metric_result.plot_data.data_in_time.freq,
                    color_options=self.color_options,
                )
            additional_graphs.append(
                AdditionalGraphInfo(
                    column_name + "_in_time",
                    {
                        "data": feature_in_time_figure["data"],
                        "layout": feature_in_time_figure["layout"],
                    },
                )
            )
            parts.append({"title": column_name + " in time", "id": column_name + "_in_time"})

        if (
            metric_result.plot_data.data_by_target is not None
            and metric_result.plot_data.data_by_target.data_for_plots is not None
        ):
            ref_data_by_target = None
            if "reference" in metric_result.plot_data.data_by_target.data_for_plots.keys():
                ref_data_by_target = metric_result.plot_data.data_by_target.data_for_plots["reference"]
            target_type = metric_result.plot_data.data_by_target.target_type
            target_name = metric_result.plot_data.data_by_target.target_name
            if column_type == "num" and target_type == "cat":
                feature_by_target_figure = plot_boxes(
                    metric_result.plot_data.data_by_target.data_for_plots["current"],
                    ref_data_by_target,
                    column_name,
                    target_name,
                    self.color_options,
                )
            if column_type == "cat" and target_type == "num":
                feature_by_target_figure = plot_boxes(
                    metric_result.plot_data.data_by_target.data_for_plots["current"],
                    ref_data_by_target,
                    target_name,
                    column_name,
                    self.color_options,
                )
            if column_type == "num" and target_type == "num":
                feature_by_target_figure = plot_num_num_rel(
                    metric_result.plot_data.data_by_target.data_for_plots["current"],
                    ref_data_by_target,
                    target_name,
                    column_name,
                    color_options=self.color_options,
                )
            if column_type == "cat" and target_type == "cat":
                feature_by_target_figure = plot_cat_cat_rel(
                    metric_result.plot_data.data_by_target.data_for_plots["current"],
                    ref_data_by_target,
                    target_name,
                    column_name,
                    color_options=self.color_options,
                )

            additional_graphs.append(
                AdditionalGraphInfo(
                    column_name + "_by_target",
                    {
                        "data": feature_by_target_figure["data"],
                        "layout": feature_by_target_figure["layout"],
                    },
                )
            )
            parts.append({"title": column_name + " by target", "id": column_name + "_by_target"})
        if column_type == "text":
            wi = BaseWidgetInfo(
                type="rich_data",
                title="",
                size=2,
                params={
                    "header": metric_result.column_name,
                    "description": column_type,
                    "metricsValuesHeaders": metrics_values_headers,
                    "metrics": self._metrics_fot_table(column_type, metric_result),
                    # "graph": graph,
                    "details": {"parts": parts, "insights": []},
                },
                additionalGraphs=additional_graphs,
            )
        else:
            wi = BaseWidgetInfo(
                type="rich_data",
                title="",
                size=2,
                params={
                    "header": metric_result.column_name,
                    "description": column_type,
                    "metricsValuesHeaders": metrics_values_headers,
                    "metrics": self._metrics_fot_table(column_type, metric_result),
                    "graph": graph,
                    "details": {"parts": parts, "insights": []},
                },
                additionalGraphs=additional_graphs,
            )
        return [wi]

    @staticmethod
    def _get_stats_with_names(
        stats_list: List[Tuple[str, str, Optional[str]]],
        current_stats: ColumnCharacteristics,
        reference_stats: Optional[ColumnCharacteristics],
    ) -> List[dict]:
        def get_values_as_string(stats_dict, field_name, field_percentage_name) -> str:
            field_value = stats_dict[field_name]

            if field_value is None:
                field_value = ""

            if field_percentage_name is None:
                return str(field_value)

            else:
                return f"{field_value} ({stats_dict[field_percentage_name]}%)"

        result = []

        current_stats_dict = {
            field.name: getattr(current_stats, field.name) for field in dataclasses.fields(current_stats)
        }

        if reference_stats is None:
            reference_stats_dict = None

        else:
            reference_stats_dict = {
                field.name: getattr(reference_stats, field.name) for field in dataclasses.fields(reference_stats)
            }

        for stat_label, stat_field, stat_field_percentage in stats_list:
            values = [get_values_as_string(current_stats_dict, stat_field, stat_field_percentage)]

            if reference_stats_dict is not None:
                values.append(get_values_as_string(reference_stats_dict, stat_field, stat_field_percentage))

            result.append(
                {
                    "label": stat_label,
                    "values": values,
                }
            )
        return result

    def _metrics_fot_table(self, column_type: str, data_quality_results: ColumnSummary):
        current_stats = data_quality_results.current_characteristics

        reference_stats = None

        if data_quality_results.reference_characteristics is not None:
            reference_stats = data_quality_results.reference_characteristics

        metrics = []
        if column_type == "cat":
            # mapping for category stats: (label, field_name_for_main_value, field_name_for_percentage)
            cat_features = [
                ("count", "count", None),
                ("unique", "unique", "unique_percentage"),
                ("most common", "most_common", "most_common_percentage"),
                ("missing", "missing", "missing_percentage"),
            ]

            if reference_stats:
                cat_features.append(("new categories", "new_in_current_values_count", None))
                cat_features.append(("missing categories", "unused_in_current_values_count", None))

            metrics.extend(self._get_stats_with_names(cat_features, current_stats, reference_stats))

        elif column_type == "num":
            # mapping for num stats: (label, field_name_for_main_value, field_name_for_percentage)
            num_features = [
                ("count", "count", None),
                ("mean", "mean", None),
                ("std", "std", None),
                ("min", "min", None),
                ("25%", "p25", None),
                ("50%", "p50", None),
                ("75%", "p75", None),
                ("max", "max", None),
                ("unique", "unique", "unique_percentage"),
                ("most common", "most_common", "most_common_percentage"),
                ("missing", "missing", "missing_percentage"),
                ("infinite", "infinite_count", "infinite_percentage"),
            ]
            metrics.extend(self._get_stats_with_names(num_features, current_stats, reference_stats))

        elif column_type == "datetime":
            # mapping for datetime stats: (label, field_name_for_main_value, field_name_for_percentage)
            datetime_features = [
                ("count", "count", None),
                ("unique", "unique", "unique_percentage"),
                ("most common", "most_common", "most_common_percentage"),
                ("missing", "missing", "missing_percentage"),
                ("first", "first", None),
                ("last", "last", None),
            ]
            metrics.extend(self._get_stats_with_names(datetime_features, current_stats, reference_stats))

        elif column_type == "text":
            text_features = [
                ("count", "count", None),
                ("missing", "missing", "missing_percentage"),
                ("Text Length min", "text_length_min", None),
                ("Text Length mean", "text_length_mean", None),
                ("Text Length max", "text_length_max", None),
                ("OOV % min", "oov_min", None),
                ("OOV % mean", "oov_mean", None),
                ("OOV % max", "oov_max", None),
                ("Non Letter Character % min", "non_letter_char_min", None),
                ("Non Letter Character % mean", "non_letter_char_mean", None),
                ("Non Letter Character % max", "non_letter_char_max", None),
            ]
            metrics.extend(self._get_stats_with_names(text_features, current_stats, reference_stats))

        return metrics



from datetime import datetime
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from evidently.analyzers.data_drift_analyzer import DataDriftAnalyzer
from evidently.model_profile.sections.base_profile_section import ProfileSection


class DataDriftProfileSection(ProfileSection):
    def part_id(self) -> str:
        return "data_drift"

    def __init__(self):
        super().__init__()
        self.analyzers_types = [DataDriftAnalyzer]
        self._result = None

    def analyzers(self):
        return self.analyzers_types

    def calculate(
        self, reference_data, current_data, column_mapping, analyzers_results
    ) -> None:
        data_drift_results = DataDriftAnalyzer.get_results(analyzers_results)
        result_json: Dict[str, Any] = data_drift_results.columns.as_dict()

        metrics_dict: Dict[str, Union[int, bool, float, Dict]] = {
            "n_features": data_drift_results.metrics.number_of_columns,
            "n_drifted_features": data_drift_results.metrics.number_of_drifted_columns,
            "share_drifted_features": data_drift_results.metrics.share_of_drifted_columns,
            "dataset_drift": data_drift_results.metrics.dataset_drift,
        }
        # add metrics to a flat dict with data drift results
        for (
            feature_name,
            feature_metrics,
        ) in data_drift_results.metrics.drift_by_columns.items():
            metrics_dict[feature_name] = {
                "current_small_hist": feature_metrics.current_small_distribution,
                "ref_small_hist": feature_metrics.reference_small_distribution,
                "feature_type": feature_metrics.column_type,
                "stattest_name": feature_metrics.stattest_name,
                "drift_score": feature_metrics.drift_score,
                "drift_detected": feature_metrics.drift_detected,
            }

        result_json["options"] = data_drift_results.options.as_dict()
        result_json["metrics"] = metrics_dict

        self._result = {
            "name": self.part_id(),
            "datetime": str(datetime.now()),
            "data": result_json,
        }

    def get_results(self) -> Optional[Dict[str, Union[str, Dict]]]:
        return self._result



"""Energy-distance test of two samples.

Name: "ed"

Import:

    >>> from evidently.calculations.stattests import energy_dist_test

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import energy_dist_test
    >>> options = DataDriftOptions(all_features_stattest=energy_dist_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="ed")
"""
from typing import Tuple

import pandas as pd
from scipy.stats import energy_distance

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _energy_dist(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    """Run the energy_distance test of two samples.
    Args:
        reference_data: reference data
        current_data: current data
        threshold: all values above this threshold propose a data drift
    Returns:
        distance: energy distance
        test_result: whether the drift is detected
    """
    distance = energy_distance(reference_data, current_data)
    return distance, distance > threshold


energy_dist_test = StatTest(
    name="ed", display_name="Energy-distance", func=_energy_dist, allowed_feature_types=["num"], default_threshold=0.1
)

register_stattest(energy_dist_test)



from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class TextWidget(Widget):
    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        return BaseWidgetInfo(
            title="Some title",
            type="text",
            size=2,
            params={
                "text": """
# Some header
Hello, world!

## Sub header
- point 1
- point 2
- point 3
                """
            },
        )

    def analyzers(self):
        return []



import dataclasses
from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text


@dataclasses.dataclass
class DataQualityStabilityMetricResult:
    number_not_stable_target: Optional[int] = None
    number_not_stable_prediction: Optional[int] = None


class DataQualityStabilityMetric(Metric[DataQualityStabilityMetricResult]):
    """Calculates stability by target and prediction"""

    def calculate(self, data: InputData) -> DataQualityStabilityMetricResult:
        result = DataQualityStabilityMetricResult()
        target_name = data.column_mapping.target
        prediction_name = data.column_mapping.prediction
        columns = [column for column in data.current_data.columns if column not in (target_name, prediction_name)]

        if not columns:
            result.number_not_stable_target = 0
            result.number_not_stable_prediction = 0
            return result

        duplicates = data.current_data[data.current_data.duplicated(subset=columns, keep=False)]

        if target_name in data.current_data:
            result.number_not_stable_target = duplicates.drop(
                data.current_data[data.current_data.duplicated(subset=columns + [target_name], keep=False)].index
            ).shape[0]

        if prediction_name in data.current_data:
            result.number_not_stable_prediction = duplicates.drop(
                data.current_data[data.current_data.duplicated(subset=columns + [prediction_name], keep=False)].index
            ).shape[0]

        return result


@default_renderer(wrap_type=DataQualityStabilityMetric)
class DataQualityStabilityMetricRenderer(MetricRenderer):
    def render_json(self, obj: DataQualityStabilityMetric) -> dict:
        return dataclasses.asdict(obj.get_result())

    def render_html(self, obj: DataQualityStabilityMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        result = [
            header_text(label="Data Stability Metrics"),
            counter(
                counters=[
                    CounterData("Not stable target", str(metric_result.number_not_stable_target)),
                    CounterData("Not stable prediction", str(metric_result.number_not_stable_prediction)),
                ]
            ),
        ]
        return result



from dataclasses import dataclass
from typing import Optional
from typing import Sequence

RED = "#ed0400"
GREY = "#4d4d4d"
COLOR_DISCRETE_SEQUENCE = (
    "#ed0400",
    "#0a5f38",
    "#6c3461",
    "#71aa34",
    "#d8dcd6",
    "#6b8ba4",
)


@dataclass
class ColorOptions:
    """Collection of colors for data visualization

    - primary_color - basic color for data visualization.
        Uses by default for all bars and lines for widgets with one dataset and as a default for current data.
    - secondary_color - basic color for second data visualization if we have two data sets, for example, reference data.
    - current_data_color - color for all current data, by default primary color is used
    - reference_data_color - color for reference data, by default secondary color is used
    - color_sequence - set of colors for drawing a number of lines in one graph, in for data quality, for example
    - fill_color - fill color for areas in line graphs
    - zero_line_color - color for base, zero line in line graphs
    - non_visible_color - color for technical, not visible dots or points for better scalability
    - underestimation_color - color for underestimation line in regression
    - overestimation_color - color for overestimation line in regression
    - majority_color - color for majority line in regression
    - lines - color for vertical lines
    - heatmap_colors - colors for heatmap
    """

    primary_color: str = RED
    secondary_color: str = GREY
    current_data_color: Optional[str] = None
    reference_data_color: Optional[str] = None
    color_sequence: Sequence[str] = COLOR_DISCRETE_SEQUENCE
    fill_color: str = "LightGreen"
    zero_line_color: str = "green"
    non_visible_color: str = "white"
    underestimation_color: str = "#6574f7"
    overestimation_color: str = "#ee5540"
    majority_color: str = "#1acc98"
    vertical_lines: str = "green"
    heatmap: str = "RdBu_r"

    def get_current_data_color(self):
        return self.current_data_color or self.primary_color

    def get_reference_data_color(self):
        return self.reference_data_color or self.secondary_color


SOLARIZED_COLOR_OPTIONS = ColorOptions(
    primary_color="#268bd2",
    secondary_color="#073642",
    current_data_color="#268bd2",
    reference_data_color="#073642",
    color_sequence=(
        "#268bd2",
        "#2aa198",
        "#859900",
        "#b58900",
        "#cb4b16",
        "#dc322f",
    ),
)

KARACHI_SUNRISE_COLOR_OPTIONS = ColorOptions(
    primary_color="#000000",
    secondary_color="#14213d",
    current_data_color="#fca311",
    reference_data_color="#e5e5e5",
    color_sequence=(
        "#dad7cd",
        "#a3b18a",
        "#588157",
        "#3a5a40",
        "#344e41",
    ),
)

BERLIN_AUTUMN_COLOR_OPTIONS = ColorOptions(
    primary_color="#3d348b",
    secondary_color="#7678ed",
    current_data_color="#f7b801",
    reference_data_color="#f18701",
    color_sequence=(
        "#f35b04",
        "#4e598c",
        "#f9c784",
        "#fcaf58",
        "#ff8c42",
    ),
)

NIGHTOWL_COLOR_OPTIONS = ColorOptions(
    primary_color="#003049",
    secondary_color="#d62828",
    current_data_color="#f77f00",
    reference_data_color="#fcbf49",
    color_sequence=(
        "#eae2b7",
        "#f08080",
        "#84a59d",
        "#f28482",
        "#f6bd60",
    ),
)



from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import Alert
from evidently.model.widget import AlertStats
from evidently.model.widget import BaseWidgetInfo
from evidently.model.widget import Insight
from evidently.model.widget import TriggeredAlertStats


class PercentWidget(Widget):
    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        return BaseWidgetInfo(
            type="percent",
            title="Example Percent Widget",
            size=1,
            details="Some additional information",
            alertsPosition="row",
            alertStats=AlertStats(
                active=4,
                triggered=TriggeredAlertStats(
                    last_24h=3,
                    period=1,
                ),
            ),
            alerts=[
                Alert(
                    value=5,
                    state="warning",
                    text="short text",
                    longText="some long description of alert",
                )
            ],
            params={"value": 40, "maxValue": 134, "details": "Some information"},
            insights=[
                Insight("Info insight", "info", "Example insight information"),
                Insight("Warning insight", "warning", "Example insight information"),
                Insight("Error insight", "error", "Example insight information"),
                Insight("Success insight", "success", "Example insight information"),
            ],
        )

    def analyzers(self):
        return []



#!/usr/bin/env python
# coding: utf-8

from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.prob_class_conf_matrix_widget import (
    ProbClassConfMatrixWidget,
)
from evidently.dashboard.widgets.prob_class_confusion_based_feature_distr_table_widget import (
    ProbClassConfusionBasedFeatureDistrTable,
)
from evidently.dashboard.widgets.prob_class_metrics_matrix_widget import (
    ProbClassMetricsMatrixWidget,
)
from evidently.dashboard.widgets.prob_class_pr_curve_widget import (
    ProbClassPRCurveWidget,
)
from evidently.dashboard.widgets.prob_class_pr_table_widget import (
    ProbClassPRTableWidget,
)
from evidently.dashboard.widgets.prob_class_pred_distr_widget import (
    ProbClassPredDistrWidget,
)
from evidently.dashboard.widgets.prob_class_prediction_cloud_widget import (
    ProbClassPredictionCloudWidget,
)
from evidently.dashboard.widgets.prob_class_quality_metrics_bar_widget import (
    ProbClassQualityMetricBarWidget,
)
from evidently.dashboard.widgets.prob_class_roc_curve_widget import (
    ProbClassRocCurveWidget,
)
from evidently.dashboard.widgets.prob_class_support_widget import ProbClassSupportWidget
from evidently.dashboard.widgets.target_name_widget import TargetNameWidget


class ProbClassificationPerformanceTab(Tab):
    widgets = [
        (
            TargetNameWidget(
                "Probabilistic Classification Model Performance Report.",
                kind="prob_classification",
            ),
            Verbose.ALWAYS,
        ),
        (
            ProbClassQualityMetricBarWidget(
                "Reference: Model Quality With Macro-average Metrics"
            ),
            Verbose.ALWAYS,
        ),
        (
            ProbClassQualityMetricBarWidget(
                "Current: Model Quality With Macro-average Metrics", "current"
            ),
            Verbose.ALWAYS,
        ),
        (ProbClassSupportWidget("Reference: Class Representation"), Verbose.ALWAYS),
        (
            ProbClassSupportWidget("Current: Class Representation", "current"),
            Verbose.ALWAYS,
        ),
        (ProbClassConfMatrixWidget("Reference: Confusion Matrix"), Verbose.FULL),
        (
            ProbClassConfMatrixWidget("Current: Confusion Matrix", "current"),
            Verbose.FULL,
        ),
        (
            ProbClassMetricsMatrixWidget("Reference: Quality Metrics by Class"),
            Verbose.ALWAYS,
        ),
        (
            ProbClassMetricsMatrixWidget(
                "Current: Quality Metrics by Class", "current"
            ),
            Verbose.ALWAYS,
        ),
        (
            ProbClassPredictionCloudWidget("Reference: Class Separation Quality"),
            Verbose.FULL,
        ),
        (
            ProbClassPredictionCloudWidget(
                "Current: Class Separation Quality", "current"
            ),
            Verbose.FULL,
        ),
        (ProbClassPredDistrWidget("Reference: Probability Distribution"), Verbose.FULL),
        (
            ProbClassPredDistrWidget("Current: Probability Distribution", "current"),
            Verbose.FULL,
        ),
        (ProbClassRocCurveWidget("Reference: ROC Curve"), Verbose.FULL),
        (ProbClassRocCurveWidget("Current: ROC Curve", "current"), Verbose.FULL),
        (ProbClassPRCurveWidget("Reference: Precision-Recall Curve"), Verbose.FULL),
        (
            ProbClassPRCurveWidget("Current: Precision-Recall Curve", "current"),
            Verbose.FULL,
        ),
        (ProbClassPRTableWidget("Reference: Precision-Recall Table"), Verbose.FULL),
        (
            ProbClassPRTableWidget("Current: Precision-Recall Table", "current"),
            Verbose.FULL,
        ),
        (
            ProbClassConfusionBasedFeatureDistrTable(
                "Classification Quality By Feature"
            ),
            Verbose.FULL,
        ),
    ]



import abc
import copy
import dataclasses
import json
import logging
from datetime import datetime
from typing import Iterator
from typing import Optional
from typing import Tuple
from typing import Union

import pandas as pd

import evidently
from evidently.base_metric import ErrorResult
from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.options import OptionsProvider
from evidently.renderers.base_renderer import DEFAULT_RENDERERS
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import RenderersDefinitions
from evidently.renderers.base_renderer import TestRenderer
from evidently.renderers.notebook_utils import determine_template
from evidently.suite.execution_graph import ExecutionGraph
from evidently.suite.execution_graph import SimpleExecutionGraph
from evidently.tests.base_test import GroupingTypes
from evidently.tests.base_test import Test
from evidently.tests.base_test import TestResult
from evidently.utils import NumpyEncoder
from evidently.utils.dashboard import SaveMode
from evidently.utils.dashboard import SaveModeMap
from evidently.utils.dashboard import TemplateParams
from evidently.utils.dashboard import save_data_file
from evidently.utils.dashboard import save_lib_files
from evidently.utils.data_preprocessing import DataDefinition


@dataclasses.dataclass
class State:
    name: str


class States:
    Init = State("Init")
    AdditionalFeatures = State("AdditionalFeatures")
    Verified = State("Verified")
    Calculated = State("Calculated")
    Tested = State("Tested")


def find_test_renderer(obj, renderers: RenderersDefinitions) -> TestRenderer:
    predefined = renderers.typed_renderers.get(obj, None)
    if predefined:
        return predefined
    if issubclass(obj, Test) and renderers.default_html_test_renderer:
        return renderers.default_html_test_renderer
    raise KeyError(f"No renderer found for {obj}")


def find_metric_renderer(obj, renderers: RenderersDefinitions) -> MetricRenderer:
    predefined = renderers.typed_renderers.get(obj, None)
    if predefined:
        return predefined
    if renderers.default_html_metric_renderer:
        return renderers.default_html_metric_renderer
    raise KeyError(f"No renderer found for {obj}")


def _discover_dependencies(test: Union[Metric, Test]) -> Iterator[Tuple[str, Union[Metric, Test]]]:
    for field_name, field in test.__dict__.items():
        if issubclass(type(field), (Metric, Test)):
            yield field_name, field


@dataclasses.dataclass
class Context:
    """Pipeline execution context tracks pipeline execution and lifecycle"""

    execution_graph: Optional[ExecutionGraph]
    metrics: list
    tests: list
    metric_results: dict
    test_results: dict
    state: State
    renderers: RenderersDefinitions


class ExecutionError(Exception):
    pass


class Display:
    # collection of all possible common options
    options_provider: OptionsProvider

    def __init__(self, options: Optional[list] = None):
        if options is None:
            options = []

        self.options_provider = OptionsProvider()

        for option in options:
            self.options_provider.add(option)

    @abc.abstractmethod
    def _build_dashboard_info(self):
        raise NotImplementedError()

    def _repr_html_(self):
        dashboard_id, dashboard_info, graphs = self._build_dashboard_info()
        template_params = TemplateParams(
            dashboard_id=dashboard_id, dashboard_info=dashboard_info, additional_graphs=graphs
        )
        return self._render(determine_template("auto"), template_params)

    def show(self, mode="auto"):
        dashboard_id, dashboard_info, graphs = self._build_dashboard_info()
        template_params = TemplateParams(
            dashboard_id=dashboard_id, dashboard_info=dashboard_info, additional_graphs=graphs
        )
        # pylint: disable=import-outside-toplevel
        try:
            from IPython.display import HTML

            return HTML(self._render(determine_template(mode), template_params))
        except ImportError as err:
            raise Exception("Cannot import HTML from IPython.display, no way to show html") from err

    def save_html(self, filename: str, mode: Union[str, SaveMode] = SaveMode.SINGLE_FILE):
        dashboard_id, dashboard_info, graphs = self._build_dashboard_info()
        if isinstance(mode, str):
            _mode = SaveModeMap.get(mode)
            if _mode is None:
                raise ValueError(f"Unexpected save mode {mode}. Expected [{','.join(SaveModeMap.keys())}]")
            mode = _mode
        if mode == SaveMode.SINGLE_FILE:
            template_params = TemplateParams(
                dashboard_id=dashboard_id,
                dashboard_info=dashboard_info,
                additional_graphs=graphs,
            )
            with open(filename, "w", encoding="utf-8") as out_file:
                out_file.write(self._render(determine_template("inline"), template_params))
        else:
            font_file, lib_file = save_lib_files(filename, mode)
            data_file = save_data_file(filename, mode, dashboard_id, dashboard_info, graphs)
            template_params = TemplateParams(
                dashboard_id=dashboard_id,
                dashboard_info=dashboard_info,
                additional_graphs=graphs,
                embed_lib=False,
                embed_data=False,
                embed_font=False,
                font_file=font_file,
                include_js_files=[lib_file, data_file],
            )
            with open(filename, "w", encoding="utf-8") as out_file:
                out_file.write(self._render(determine_template("inline"), template_params))

    @abc.abstractmethod
    def as_dict(self) -> dict:
        raise NotImplementedError()

    def _get_json_content(self) -> dict:
        """Return all data for json representation"""
        result = {
            "version": evidently.__version__,
            "timestamp": str(datetime.now()),
        }
        result.update(self.as_dict())
        return result

    def json(self) -> str:
        return json.dumps(
            self._get_json_content(),
            cls=NumpyEncoder,
        )

    def save_json(self, filename):
        with open(filename, "w", encoding="utf-8") as out_file:
            json.dump(self._get_json_content(), out_file, cls=NumpyEncoder)

    def _render(self, temple_func, template_params: TemplateParams):
        return temple_func(params=template_params)


class Suite:
    context: Context

    def __init__(self):
        self.context = Context(
            execution_graph=None,
            metrics=[],
            tests=[],
            metric_results={},
            test_results={},
            state=States.Init,
            renderers=DEFAULT_RENDERERS,
        )

    def add_test(self, test: Test):
        test.set_context(self.context)
        for field_name, dependency in _discover_dependencies(test):
            if isinstance(dependency, Metric):
                self.add_metric(dependency)

            if isinstance(dependency, Test):
                dependency_copy = copy.copy(dependency)
                test.__setattr__(field_name, dependency_copy)
                self.add_test(dependency_copy)
        self.context.tests.append(test)
        self.context.state = States.Init

    def add_metric(self, metric: Metric):
        metric.set_context(self.context)

        for field_name, dependency in _discover_dependencies(metric):
            if isinstance(dependency, Metric):
                self.add_metric(dependency)

            if isinstance(dependency, Test):
                dependency_copy = copy.copy(dependency)
                metric.__setattr__(field_name, dependency_copy)
                self.add_test(dependency_copy)
        self.context.metrics.append(metric)
        self.context.state = States.Init

    def verify(self):
        self.context.execution_graph = SimpleExecutionGraph(self.context.metrics, self.context.tests)
        self.context.state = States.Verified

    def create_additional_features(
        self,
        current_data: pd.DataFrame,
        reference_data: Optional[pd.DataFrame],
        data_definition: DataDefinition,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        curr_additional_data = None
        ref_additional_data = None
        features = {}
        if self.context.execution_graph is not None:
            execution_graph: ExecutionGraph = self.context.execution_graph
            for metric, calculation in execution_graph.get_metric_execution_iterator():
                try:
                    required_features = metric.required_features(data_definition)
                except Exception as e:
                    logging.error(f"failed to get features for {type(metric)}: {e}")
                    continue
                for feature in required_features:
                    params = feature.get_parameters()
                    if params is not None:
                        _id = (type(feature), params)
                        if _id in features:
                            continue
                        features[_id] = feature
                    feature_data = feature.generate_feature(current_data, data_definition)
                    feature_data.columns = [f"{feature.__class__.__name__}.{old}" for old in feature_data.columns]
                    if curr_additional_data is None:
                        curr_additional_data = feature_data
                    else:
                        curr_additional_data = curr_additional_data.join(feature_data)
                    if reference_data is None:
                        continue
                    ref_feature_data = feature.generate_feature(reference_data, data_definition)
                    ref_feature_data.columns = [
                        f"{feature.__class__.__name__}.{old}" for old in ref_feature_data.columns
                    ]

                    if ref_additional_data is None:
                        ref_additional_data = ref_feature_data
                    else:
                        ref_additional_data = ref_additional_data.join(ref_feature_data)
        return curr_additional_data, ref_additional_data

    def run_calculate(self, data: InputData):
        if self.context.state in [States.Init]:
            self.verify()

        if self.context.state in [States.Calculated, States.Tested]:
            return

        if self.context.execution_graph is not None:
            execution_graph: ExecutionGraph = self.context.execution_graph

            calculations = {}
            for metric, calculation in execution_graph.get_metric_execution_iterator():
                if calculation not in calculations:
                    logging.debug(f"Executing {type(calculation)}...")
                    try:
                        calculations[calculation] = calculation.calculate(data)
                    except BaseException as ex:
                        calculations[calculation] = ErrorResult(ex)
                else:
                    logging.debug(f"Using cached result for {type(calculation)}")
                self.context.metric_results[metric] = calculations[calculation]

        self.context.state = States.Calculated

    def run_checks(self):
        if self.context.state in [States.Init, States.Verified]:
            raise ExecutionError("No calculation was made, run 'run_calculate' first'")

        test_results = {}

        for test in self.context.execution_graph.get_test_execution_iterator():
            try:
                logging.debug(f"Executing {type(test)}...")
                test_results[test] = test.check()
            except BaseException as ex:
                test_results[test] = TestResult(
                    name=test.name, status=TestResult.ERROR, description=f"Test failed with exceptions: {ex}"
                )
            test_results[test].groups.update(
                {
                    GroupingTypes.TestGroup.id: test.group,
                    GroupingTypes.TestType.id: test.name,
                }
            )

        self.context.test_results = test_results
        self.context.state = States.Tested



import dataclasses
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_absolute_percentage_error
from sklearn.metrics import mean_squared_error

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.metrics.regression_performance.regression_quality import RegressionQualityMetric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import table_data
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class RegressionDummyMetricResults:
    rmse_default: float
    mean_abs_error_default: float
    mean_abs_perc_error_default: float
    abs_error_max_default: float
    mean_abs_error_by_ref: Optional[float] = None
    mean_abs_error: Optional[float] = None
    mean_abs_perc_error_by_ref: Optional[float] = None
    mean_abs_perc_error: Optional[float] = None
    rmse_by_ref: Optional[float] = None
    rmse: Optional[float] = None
    abs_error_max_by_ref: Optional[float] = None
    abs_error_max: Optional[float] = None


class RegressionDummyMetric(Metric[RegressionDummyMetricResults]):
    quality_metric: RegressionQualityMetric

    def __init__(self):
        self.quality_metric = RegressionQualityMetric()

    def calculate(self, data: InputData) -> RegressionDummyMetricResults:
        quality_metric: Optional[RegressionQualityMetric]
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction

        if target_name is None:
            raise ValueError("The column 'target' should present")
        if prediction_name is None:
            quality_metric = None
        else:
            quality_metric = self.quality_metric
        if prediction_name is not None and not isinstance(prediction_name, str):
            raise ValueError("Expect one column for prediction. List of columns was provided.")

        # dummy by current
        # mae
        dummy_preds = data.current_data[target_name].median()
        mean_abs_error_default = mean_absolute_error(
            y_true=data.current_data[target_name], y_pred=[dummy_preds] * data.current_data.shape[0]
        )
        # rmse
        dummy_preds = data.current_data[target_name].mean()
        rmse_default = mean_squared_error(
            y_true=data.current_data[target_name], y_pred=[dummy_preds] * data.current_data.shape[0]
        )
        # mape default values
        # optimal constant for mape
        s = data.current_data[target_name]
        inv_y = 1 / s[s != 0].values
        w = inv_y / sum(inv_y)
        idxs = np.argsort(w)
        sorted_w = w[idxs]
        sorted_w_cumsum = np.cumsum(sorted_w)
        idx = np.where(sorted_w_cumsum > 0.5)[0][0]
        pos = idxs[idx]
        dummy_preds = s[s != 0].values[pos]

        mean_abs_perc_error_default = (
            mean_absolute_percentage_error(
                y_true=data.current_data[target_name],
                y_pred=[dummy_preds] * data.current_data.shape[0],
            )
            * 100
        )
        # max error default values
        y_true = data.current_data[target_name]
        abs_error_max_default = np.abs(y_true - y_true.median()).max()

        # dummy by reference
        mean_abs_error_by_ref: Optional[float] = None
        mean_abs_perc_error_by_ref: Optional[float] = None
        rmse_by_ref: Optional[float] = None
        abs_error_max_by_ref: Optional[float] = None
        if data.reference_data is not None:
            # mae
            dummy_preds = data.reference_data[target_name].median()
            mean_abs_error_by_ref = mean_absolute_error(
                y_true=data.current_data[target_name], y_pred=[dummy_preds] * data.current_data.shape[0]
            )
            # rmse
            dummy_preds = data.reference_data[target_name].mean()
            rmse_by_ref = mean_squared_error(
                y_true=data.current_data[target_name], y_pred=[dummy_preds] * data.current_data.shape[0]
            )
            # mape default values
            # optimal constant for mape
            s = data.reference_data[target_name]
            inv_y = 1 / s[s != 0].values
            w = inv_y / sum(inv_y)
            idxs = np.argsort(w)
            sorted_w = w[idxs]
            sorted_w_cumsum = np.cumsum(sorted_w)
            idx = np.where(sorted_w_cumsum > 0.5)[0][0]
            pos = idxs[idx]
            dummy_preds = s[s != 0].values[pos]

            mean_abs_perc_error_by_ref = (
                mean_absolute_percentage_error(
                    y_true=data.current_data[target_name],
                    y_pred=[dummy_preds] * data.current_data.shape[0],
                )
                * 100
            )
            # max error default values
            y_true = data.current_data[target_name]
            y_pred = data.reference_data[target_name].median()
            abs_error_max_by_ref = np.abs(y_true - y_pred).max()

        return RegressionDummyMetricResults(
            rmse_default=rmse_default,
            mean_abs_error_default=mean_abs_error_default,
            mean_abs_perc_error_default=mean_abs_perc_error_default,
            abs_error_max_default=abs_error_max_default,
            mean_abs_error_by_ref=mean_abs_error_by_ref,
            mean_abs_error=(quality_metric.get_result().mean_abs_error if quality_metric is not None else None),
            mean_abs_perc_error_by_ref=mean_abs_perc_error_by_ref,
            mean_abs_perc_error=(
                quality_metric.get_result().mean_abs_perc_error if quality_metric is not None else None
            ),
            rmse_by_ref=rmse_by_ref,
            rmse=quality_metric.get_result().rmse if quality_metric is not None else None,
            abs_error_max_by_ref=abs_error_max_by_ref,
            abs_error_max=(quality_metric.get_result().abs_error_max if quality_metric is not None else None),
        )


@default_renderer(wrap_type=RegressionDummyMetric)
class RegressionDummyMetricRenderer(MetricRenderer):
    def render_json(self, obj: RegressionDummyMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    def render_html(self, obj: RegressionDummyMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        in_table_data = pd.DataFrame(data=["MAE", "RMSE", "MAPE", "MAX_ERROR"])
        columns = ["Metric"]
        if (
            metric_result.abs_error_max_by_ref is not None
            and metric_result.mean_abs_perc_error_by_ref is not None
            and metric_result.rmse_by_ref is not None
            and metric_result.mean_abs_error_by_ref is not None
        ):
            in_table_data["by_ref"] = [
                metric_result.mean_abs_error_by_ref,
                metric_result.rmse_by_ref,
                metric_result.mean_abs_perc_error_by_ref,
                metric_result.abs_error_max_by_ref,
            ]
            columns.append("Dummy (by rerefence)")
        in_table_data["by_carr"] = [
            metric_result.mean_abs_error_default,
            metric_result.rmse_default,
            metric_result.mean_abs_perc_error_default,
            metric_result.abs_error_max_default,
        ]
        if "Dummy (by rerefence)" in columns:
            columns.append("Dummy (by current)")
        else:
            columns.append("Dummy")
        if (
            metric_result.mean_abs_error is not None
            and metric_result.rmse is not None
            and metric_result.mean_abs_perc_error is not None
            and metric_result.abs_error_max is not None
        ):
            in_table_data["model_quality"] = [
                metric_result.mean_abs_error,
                metric_result.rmse,
                metric_result.mean_abs_perc_error,
                metric_result.abs_error_max,
            ]
            columns.append("Model")

        return [
            header_text(label="Dummy Regreesion Quality"),
            table_data(column_names=columns, data=np.around(in_table_data, 3).values, title=""),
        ]



import logging
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from evidently.options import DataDriftOptions
from evidently.options import QualityMetricsOptions
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.runner.loader import DataLoader
from evidently.runner.loader import DataOptions
from evidently.runner.loader import SamplingOptions


@dataclass
class RunnerOptions:
    reference_data_path: str
    reference_data_options: DataOptions
    reference_data_sampling: Optional[SamplingOptions]
    current_data_path: Optional[str]
    current_data_options: Optional[DataOptions]
    current_data_sampling: Optional[SamplingOptions]
    column_mapping: ColumnMapping
    options: List[object]
    output_path: str


options_mapping: Dict[str, Type] = {
    "data_drift": DataDriftOptions,
    "quality_metrics": QualityMetricsOptions,
}


def parse_options(raw_dict: Optional[Dict[str, Dict[str, object]]]) -> List[object]:
    if raw_dict is None:
        return []
    result = []
    for key, params in raw_dict.items():
        opt_class = options_mapping.get(key, None)
        if opt_class is None:
            raise ValueError(f"No options with id {key} exists")
        result.append(opt_class(**params))
    return result


class Runner:
    def __init__(self, options: RunnerOptions):
        self.options = options

    def _parse_data(self):
        loader = DataLoader()

        reference_data = loader.load(
            self.options.reference_data_path, self.options.reference_data_options, self.options.reference_data_sampling
        )
        logging.info(f"reference dataset loaded: {len(reference_data)} rows")
        if self.options.current_data_path:
            current_data = loader.load(
                self.options.current_data_path, self.options.current_data_options, self.options.current_data_sampling
            )
            logging.info(f"current dataset loaded: {len(current_data)} rows")
        else:
            current_data = None

        return reference_data, current_data



#!/usr/bin/env python
# coding: utf-8

from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class BarWidget(Widget):
    def analyzers(self):
        return []

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        return BaseWidgetInfo(
            type="big_graph",
            title=self.title,
            size=2,
            params={
                "data": [
                    {
                        "marker": {"color": "#ed0400"},
                        "type": "bar",
                        "x": [1, 2, 3, 4, 5],
                        "y": [2, 5, 3, 1, 2],
                    }
                ],
                "layout": {
                    "xaxis": {"title": {"text": "Features"}},
                    "yaxis": {"showticklabels": True, "title": {"text": "Correlation"}},
                },
            },
            alerts=[],
            insights=[],
            details="",
            alertsPosition="row",
            additionalGraphs=[],
        )



#!/usr/bin/env python
# coding: utf-8
import abc
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.model.widget import BaseWidgetInfo
from evidently.options import OptionsProvider


class Widget:
    title: str
    wi: Optional[BaseWidgetInfo]
    options_provider: OptionsProvider

    def __init__(self, title: str):
        self.title = title
        self.wi = None

    @abc.abstractmethod
    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        raise NotImplementedError()

    @abc.abstractmethod
    def analyzers(self):
        return []



#!/usr/bin/env python
# coding: utf-8
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.data_drift_analyzer import DataDriftAnalyzer
from evidently.calculations.data_drift import ColumnDataDriftMetrics
from evidently.dashboard.widgets.utils import CutQuantileTransformer
from evidently.dashboard.widgets.utils import fig_to_json
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions
from evidently.options import DataDriftOptions
from evidently.options import QualityMetricsOptions


def _generate_feature_params(name: str, data: ColumnDataDriftMetrics) -> dict:
    if (
        data.current_small_distribution is None
        or data.reference_small_distribution is None
    ):
        return {}

    current_small_hist = data.current_small_distribution
    ref_small_hist = data.reference_small_distribution
    feature_type = data.column_type
    drift_score = data.drift_score
    distr_sim_test = "Detected" if data.drift_detected else "Not Detected"
    parts = []

    if data.column_type == "num":
        parts.append({"title": "Data drift", "id": f"{name}_drift", "type": "widget"})

    parts.append({"title": "Data distribution", "id": f"{name}_distr"})
    return {
        "details": {"parts": parts, "insights": []},
        "f1": name,
        "f6": feature_type,
        "stattest_name": data.stattest_name,
        "f3": {"x": list(ref_small_hist[1]), "y": list(ref_small_hist[0])},
        "f4": {"x": list(current_small_hist[1]), "y": list(current_small_hist[0])},
        "f2": distr_sim_test,
        "f5": round(drift_score, 6),
    }


def _generate_additional_graph_num_feature(
    name: str,
    reference_data: pd.DataFrame,
    current_data: pd.DataFrame,
    date_column: Optional[str],
    data_drift_options: DataDriftOptions,
    quality_metrics_options: QualityMetricsOptions,
    color_options: ColorOptions,
) -> List[AdditionalGraphInfo]:
    # plot distributions
    conf_interval_n_sigmas = quality_metrics_options.conf_interval_n_sigmas
    fig = go.Figure()
    if data_drift_options.xbins and data_drift_options.xbins.get(name):
        current_xbins = data_drift_options.xbins.get(name)
        current_nbinsx = None
    else:
        current_xbins = None
        current_nbinsx = data_drift_options.get_nbinsx(name)
    quantiles = quality_metrics_options.get_cut_quantile(name)
    if quantiles:
        side, q = quantiles
        cqt = CutQuantileTransformer(side=side, q=q)
        cqt.fit(reference_data[name])
        reference_data_to_plot = cqt.transform(reference_data[name])
        current_data_to_plot = cqt.transform(current_data[name])
    else:
        reference_data_to_plot = reference_data[name]
        current_data_to_plot = current_data[name]
    fig.add_trace(
        go.Histogram(
            x=reference_data_to_plot,
            marker_color=color_options.get_reference_data_color(),
            opacity=0.6,
            xbins=current_xbins,
            nbinsx=current_nbinsx,
            name="Reference",
            histnorm="probability",
        )
    )

    fig.add_trace(
        go.Histogram(
            x=current_data_to_plot,
            marker_color=color_options.get_current_data_color(),
            opacity=0.6,
            xbins=current_xbins,
            nbinsx=current_nbinsx,
            name="Current",
            histnorm="probability",
        )
    )
    fig.update_layout(
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
        xaxis_title=name,
        yaxis_title="Share",
    )

    distr_figure = fig_to_json(fig)

    # plot drift
    reference_mean = np.mean(reference_data[name][np.isfinite(reference_data[name])])
    reference_std = np.std(
        reference_data[name][np.isfinite(reference_data[name])], ddof=1
    )
    x_title = "Timestamp" if date_column else "Index"

    fig = go.Figure()

    fig.add_trace(
        go.Scattergl(
            x=current_data[date_column] if date_column else current_data.index,
            y=current_data[name],
            mode="markers",
            name="Current",
            marker=dict(size=6, color=color_options.get_current_data_color()),
        )
    )

    if date_column:
        # take the second value because of some plotly issues
        x0 = current_data[date_column].sort_values().iloc[1]

    else:
        x0 = current_data.index.sort_values()[1]

    fig.add_trace(
        go.Scattergl(
            x=[x0, x0],
            y=[
                reference_mean - conf_interval_n_sigmas * reference_std,
                reference_mean + conf_interval_n_sigmas * reference_std,
            ],
            mode="markers",
            name="Current",
            marker=dict(
                size=0.01, color=color_options.non_visible_color, opacity=0.005
            ),
            showlegend=False,
        )
    )

    fig.update_layout(
        xaxis_title=x_title,
        yaxis_title=name,
        showlegend=True,
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
        shapes=[
            dict(
                type="rect",
                # x-reference is assigned to the x-values
                xref="paper",
                # y-reference is assigned to the plot paper [0,1]
                yref="y",
                x0=0,
                y0=reference_mean - conf_interval_n_sigmas * reference_std,
                x1=1,
                y1=reference_mean + conf_interval_n_sigmas * reference_std,
                fillcolor=color_options.fill_color,
                opacity=0.5,
                layer="below",
                line_width=0,
            ),
            dict(
                type="line",
                name="Reference",
                xref="paper",
                yref="y",
                x0=0,  # min(testset_agg_by_date.index),
                y0=reference_mean,
                x1=1,  # max(testset_agg_by_date.index),
                y1=reference_mean,
                line=dict(color=color_options.zero_line_color, width=3),
            ),
        ],
    )

    drift_figure = fig_to_json(fig)

    # add distributions data
    return [
        AdditionalGraphInfo(
            f"{name}_distr",
            {"data": distr_figure["data"], "layout": distr_figure["layout"]},
        ),
        AdditionalGraphInfo(
            f"{name}_drift",
            {
                "title": "",
                "size": 2,
                "text": "",
                "type": "big_graph",
                "params": {
                    "data": drift_figure["data"],
                    "layout": drift_figure["layout"],
                },
            },
        ),
    ]


def _generate_additional_graph_cat_feature(
    name: str,
    reference_data: pd.DataFrame,
    current_data: pd.DataFrame,
    color_options: ColorOptions,
) -> List[AdditionalGraphInfo]:
    fig = go.Figure()
    feature_ref_data = reference_data[name].dropna()
    feature_cur_data = current_data[name].dropna()
    reference_data_to_plot = list(
        reversed(list(map(list, zip(*feature_ref_data.value_counts().items()))))
    )
    current_data_to_plot = list(
        reversed(list(map(list, zip(*feature_cur_data.value_counts().items()))))
    )
    fig.add_trace(
        go.Bar(
            x=reference_data_to_plot[1],
            y=reference_data_to_plot[0],
            marker_color=color_options.get_reference_data_color(),
            opacity=0.6,
            name="Reference",
        )
    )

    fig.add_trace(
        go.Bar(
            x=current_data_to_plot[1],
            y=current_data_to_plot[0],
            marker_color=color_options.get_current_data_color(),
            opacity=0.6,
            name="Current",
        )
    )
    fig.update_layout(
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
        xaxis_title=name,
        yaxis_title="Share",
    )

    distr_figure = fig_to_json(fig)
    return [
        AdditionalGraphInfo(
            f"{name}_distr",
            {"data": distr_figure["data"], "layout": distr_figure["layout"]},
        )
    ]


class DataDriftTableWidget(Widget):
    def analyzers(self):
        return [DataDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        data_drift_results = DataDriftAnalyzer.get_results(analyzers_results)
        all_features = data_drift_results.columns.get_all_features_list()
        date_column = data_drift_results.columns.utility_columns.date
        target_column = data_drift_results.columns.utility_columns.target
        prediction_column = data_drift_results.columns.utility_columns.prediction

        if current_data is None:
            raise ValueError("current_data should be present")

        # set params data
        params_data = []
        data_drift_options = self.options_provider.get(DataDriftOptions)
        quality_metrics_options = self.options_provider.get(QualityMetricsOptions)

        # sort columns by drift score
        df_for_sort = pd.DataFrame()
        df_for_sort["features"] = all_features
        df_for_sort["scores"] = [
            data_drift_results.metrics.drift_by_columns[feature].drift_score
            for feature in all_features
        ]
        all_features = df_for_sort.sort_values(
            "scores", ascending=False
        ).features.tolist()
        columns = []

        # move target and prediction to the top of the table
        if target_column:
            columns.append(target_column)

            if target_column in all_features:
                all_features.remove(target_column)

        if prediction_column and isinstance(prediction_column, str):
            columns.append(prediction_column)

            if prediction_column in all_features:
                all_features.remove(prediction_column)

        columns = columns + all_features

        for feature_name in columns:
            params_data.append(
                _generate_feature_params(
                    feature_name,
                    data_drift_results.metrics.drift_by_columns[feature_name],
                )
            )

        # set additionalGraphs
        additional_graphs_data = []
        for feature_name in columns:
            # plot distributions
            if (
                data_drift_results.metrics.drift_by_columns[feature_name].column_type
                == "num"
            ):
                additional_graphs_data += _generate_additional_graph_num_feature(
                    feature_name,
                    reference_data,
                    current_data,
                    date_column,
                    data_drift_options,
                    quality_metrics_options,
                    color_options,
                )
            elif (
                data_drift_results.metrics.drift_by_columns[feature_name].column_type
                == "cat"
            ):
                additional_graphs_data += _generate_additional_graph_cat_feature(
                    feature_name, reference_data, current_data, color_options
                )
        n_drifted_features = data_drift_results.metrics.number_of_drifted_columns
        dataset_drift = data_drift_results.metrics.dataset_drift
        n_features = data_drift_results.metrics.number_of_columns
        drift_share = data_drift_results.metrics.share_of_drifted_columns

        title_prefix = (
            f"Drift is detected for {drift_share * 100:.2f}% of features ({n_drifted_features}"
            f" out of {n_features}). "
        )
        title_suffix = (
            "Dataset Drift is detected."
            if dataset_drift
            else "Dataset Drift is NOT detected."
        )

        return BaseWidgetInfo(
            title=title_prefix + title_suffix,
            type="big_table",
            details="",
            alerts=[],
            alertsPosition="row",
            insights=[],
            size=2,
            params={
                "rowsPerPage": min(n_features, 10),
                "columns": [
                    {"title": "Feature", "field": "f1"},
                    {"title": "Type", "field": "f6"},
                    {
                        "title": "Reference Distribution",
                        "field": "f3",
                        "type": "histogram",
                        "options": {
                            "xField": "x",
                            "yField": "y",
                            "color": color_options.primary_color,
                        },
                    },
                    {
                        "title": "Current Distribution",
                        "field": "f4",
                        "type": "histogram",
                        "options": {
                            "xField": "x",
                            "yField": "y",
                            "color": color_options.primary_color,
                        },
                    },
                    {"title": "Data Drift", "field": "f2"},
                    {"title": "Stat Test", "field": "stattest_name"},
                    {"title": "Drift Score", "field": "f5"},
                ],
                "data": params_data,
            },
            additionalGraphs=additional_graphs_data,
        )



#!/usr/bin/env python
# coding: utf-8
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.classification_performance_analyzer import (
    ClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class ClassQualityMetricsBarWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results = ClassificationPerformanceAnalyzer.get_results(analyzers_results)
        target_name = results.columns.utility_columns.target
        prediction_name = results.columns.utility_columns.prediction

        if target_name is None or prediction_name is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns."
                )
            return None

        if self.dataset == "current":
            result_metrics = results.current_metrics

        elif self.dataset == "reference":
            result_metrics = results.reference_metrics

            if not result_metrics:
                raise ValueError(
                    f"Widget [{self.title}] required 'reference' results from"
                    f" {ClassificationPerformanceAnalyzer.__name__} but no data found"
                )

        else:
            raise ValueError(
                f"Widget [{self.title}] requires 'current' or 'reference' dataset value"
            )

        if result_metrics is None:
            return None

        # plot quality metrics bar
        return BaseWidgetInfo(
            title=self.title,
            type="counter",
            size=2,
            params={
                "counters": [
                    {
                        "value": str(round(result_metrics.accuracy, 3)),
                        "label": "Accuracy",
                    },
                    {
                        "value": str(round(result_metrics.precision, 3)),
                        "label": "Precision",
                    },
                    {"value": str(round(result_metrics.recall, 3)), "label": "Recall"},
                    {"value": str(round(result_metrics.f1, 3)), "label": "F1"},
                ]
            },
        )



"""Cramer-Von-mises test of two samples.

Name: "cramer_von_mises"

Import:

    >>> from evidently.calculations.stattests import cramer_von_mises

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import cramer_von_mises
    >>> options = DataDriftOptions(all_features_stattest=cramer_von_mises)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="cramer_von_mises")
"""
from itertools import combinations
from typing import Generator
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.special import gammaln
from scipy.special import kv
from scipy.stats import rankdata

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _all_partitions(nx: int, ny: int) -> Generator[Tuple[int, int], None, None]:
    """
    Args:
        nx: int
        ny: int
    Return:
        None
    """
    z = np.arange(nx + ny)
    for c in combinations(z, nx):
        x = np.array(c)
        mask = np.ones(nx + ny, bool)
        mask[x] = False
        y = z[mask]
        yield x, y


def _pval_cvm_2samp_exact(s: float, nx: int, ny: int):
    """Compute the exact p
    Args:
        s: test statistic
        nx: sample size
        ny: sample size
    Returns:
        p_value: p_value
    """
    rangex = np.arange(nx)
    rangey = np.arange(ny)

    us = []

    for x, y in _all_partitions(nx, ny):
        u = nx * np.sum((x - rangex) ** 2)
        u += ny * np.sum((y - rangey) ** 2)
        us.append(u)

    u, cnt = np.unique(us, return_counts=True)
    p_value = np.sum(cnt[u >= s]) / np.sum(cnt)
    return p_value


class CramerVonMisesResult:
    def __init__(self, statistic, pvalue):
        self.statistic = statistic
        self.pvalue = pvalue

    def __repr__(self):
        return f"{self.__class__.__name__}(statistic={self.statistic}, " f"pvalue={self.pvalue})"


def _cdf_cvm_inf(x: float) -> float:
    """Calculate the cdf of the Cramér-von Mises statistic
    Args:
        x: float
    Returns:
        tot: float
    """
    x = np.asarray(x)

    def term(x, k):
        u = np.exp(gammaln(k + 0.5) - gammaln(k + 1)) / (np.pi**1.5 * np.sqrt(x))
        y = 4 * k + 1
        q = y**2 / (16 * x)
        b = kv(0.25, q)
        return u * np.sqrt(y) * np.exp(-q) * b

    tot = np.zeros_like(x, dtype="float")
    cond = np.ones_like(x, dtype="bool")
    k = 0
    while np.any(cond):
        z = term(x[cond], k)  # type: ignore
        tot[cond] = tot[cond] + z
        cond[cond] = np.abs(z) >= 1e-7
        k += 1

    return tot


def _cvm_2samp(x: np.ndarray, y: np.ndarray, method: str = "auto") -> CramerVonMisesResult:
    """Perform the two-sample Cramér-von Mises test
    Args:
        x : array_like
        y : array_like
        method : {'auto', 'asymptotic', 'exact'}, optional
    Returns:
        res : object with attributes
        statistic : Cramér-von Mises statistic.
        pvalue : float
    """

    xa = np.sort(np.asarray(x))
    ya = np.sort(np.asarray(y))

    if xa.size <= 1 or ya.size <= 1:
        raise ValueError("x and y must contain at least two observations.")
    if xa.ndim > 1 or ya.ndim > 1:
        raise ValueError("The samples must be one-dimensional.")
    if method not in ["auto", "exact", "asymptotic"]:
        raise ValueError("method must be either auto, exact or asymptotic.")

    nx = len(xa)
    ny = len(ya)

    if method == "auto":
        if max(nx, ny) > 10:
            method = "asymptotic"
        else:
            method = "exact"

    z = np.concatenate([xa, ya])
    r = rankdata(z, method="average")

    rx = r[:nx]
    ry = r[nx:]

    u = nx * np.sum((rx - np.arange(1, nx + 1)) ** 2)

    u += ny * np.sum((ry - np.arange(1, ny + 1)) ** 2)

    k, N = nx * ny, nx + ny
    t = u / (k * N) - (4 * k - 1) / (6 * N)

    if method == "exact":
        p = _pval_cvm_2samp_exact(u, nx, ny)
    else:
        et = (1 + 1 / N) / 6

        vt = (N + 1) * (4 * k * N - 3 * (nx**2 + ny**2) - 2 * k)
        vt = vt / (45 * N**2 * 4 * k)  # type: ignore

        tn = 1 / 6 + (t - et) / np.sqrt(45 * vt)

        if tn < 0.003:
            p = 1.0
        else:
            p = max(0, 1.0 - _cdf_cvm_inf(tn))

    return CramerVonMisesResult(statistic=t, pvalue=p)


def _cramer_von_mises(
    reference_data: pd.Series,
    current_data: pd.Series,
    feature_type: str,
    threshold: float,
) -> Tuple[float, bool]:
    """Run the two-sample Cramer-Von-mises test of two samples.
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: level of significance
    Returns:
        p_value: p-value
        test_result: whether the drift is detected
    """
    res = _cvm_2samp(reference_data, current_data)
    return res.pvalue, res.pvalue <= threshold


cramer_von_mises = StatTest(
    name="cramer_von_mises",
    display_name="Cramer-Von-mises",
    func=_cramer_von_mises,
    allowed_feature_types=["num"],
    default_threshold=0.1,
)

register_stattest(cramer_von_mises)



"""Mann-Whitney U-rank test of two samples.

Name: "mannw"

Import:

    >>> from evidently.calculations.stattests import mann_whitney_u_stat_test

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import mann_whitney_u_stat_test
    >>> options = DataDriftOptions(all_features_stattest=mann_whitney_u_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="mannw")
"""
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.stats import norm

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest
from evidently.calculations.stattests.utils import get_unique_not_nan_values_list_from_series


def proportions_diff_z_stat_ind(ref: pd.DataFrame, curr: pd.DataFrame):
    # pylint: disable=invalid-name
    n1 = len(ref)
    n2 = len(curr)

    p1 = float(sum(ref)) / n1
    p2 = float(sum(curr)) / n2
    P = float(p1 * n1 + p2 * n2) / (n1 + n2)

    return (p1 - p2) / np.sqrt(P * (1 - P) * (1.0 / n1 + 1.0 / n2))


def proportions_diff_z_test(z_stat, alternative="two-sided"):
    if alternative == "two-sided":
        return 2 * (1 - norm.cdf(np.abs(z_stat)))

    if alternative == "less":
        return norm.cdf(z_stat)

    if alternative == "greater":
        return 1 - norm.cdf(z_stat)

    raise ValueError("alternative not recognized\n" "should be 'two-sided', 'less' or 'greater'")


def _z_stat_test(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    """Compute the Z test between two arrays
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
    Returns:
        pvalue: the two-tailed p-value for the test
        test_result: whether the drift is detected
    """
    if (
        reference_data.nunique() == 1
        and current_data.nunique() == 1
        and reference_data.unique()[0] == current_data.unique()[0]
    ):
        p_value = 1
    else:
        keys = sorted(
            get_unique_not_nan_values_list_from_series(current_data=current_data, reference_data=reference_data)
        )
        p_value = proportions_diff_z_test(
            proportions_diff_z_stat_ind(
                reference_data.apply(lambda x, key=keys[0]: 0 if x == key else 1),
                current_data.apply(lambda x, key=keys[0]: 0 if x == key else 1),
            )
        )
    return p_value, p_value < threshold


z_stat_test = StatTest(
    name="z",
    display_name="Z-test p_value",
    func=_z_stat_test,
    allowed_feature_types=["cat"],
)

register_stattest(z_stat_test)



"""Predefined Test Presets for Test Suite"""
from .classification_binary import BinaryClassificationTestPreset
from .classification_binary_topk import BinaryClassificationTopKTestPreset
from .classification_multiclass import MulticlassClassificationTestPreset
from .data_drift import DataDriftTestPreset
from .data_quality import DataQualityTestPreset
from .data_stability import DataStabilityTestPreset
from .no_target_performance import NoTargetPerformanceTestPreset
from .regression import RegressionTestPreset

__all__ = [
    BinaryClassificationTestPreset.__name__,
    BinaryClassificationTopKTestPreset.__name__,
    MulticlassClassificationTestPreset.__name__,
    DataDriftTestPreset.__name__,
    DataQualityTestPreset.__name__,
    DataStabilityTestPreset.__name__,
    NoTargetPerformanceTestPreset.__name__,
    RegressionTestPreset.__name__,
]



import pandas as pd

from evidently.base_metric import ColumnName
from evidently.base_metric import additional_feature
from evidently.features.generated_features import GeneratedFeature
from evidently.utils.data_preprocessing import DataDefinition


class NonLetterCharacterPercentage(GeneratedFeature):
    def __init__(self, column_name: str):
        self.column_name = column_name

    def generate_feature(self, data: pd.DataFrame, data_definition: DataDefinition) -> pd.DataFrame:

        # counts share of characters that are not letters or spaces
        def non_letter_share(s):
            if s is None:
                return 0
            non_letters_num = 0
            for ch in s:
                if not ch.isalpha() and ch != " ":
                    non_letters_num += 1
            return 100 * non_letters_num / len(s)

        return pd.DataFrame(dict([(self.column_name, data[self.column_name].apply(non_letter_share))]))

    def feature_name(self) -> ColumnName:
        return additional_feature(self, self.column_name)



#!/usr/bin/env python
# coding: utf-8
from typing import Optional

import pandas as pd
from plotly import graph_objs as go

from evidently import ColumnMapping
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.model.widget import Insight


class ExpandableListWidget(Widget):
    def analyzers(self):
        return []

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        fig = go.Figure()
        fig.add_trace(go.Bar(x=[1, 2, 3, 4, 5], y=[0.1, 0.2, 0.3, 0.4, 0.5]))

        fig_dict = fig.to_dict()

        return BaseWidgetInfo(
            title="",
            size=2,
            type="list",
            widgets=[
                BaseWidgetInfo(
                    type="rich_data",
                    title="",
                    size=2,
                    params={
                        "header": "Header",
                        "description": "optional description",
                        "metricsValuesHeaders": ["reference", "current"],
                        "metrics": [
                            {"label": "f1", "values": [0.3, 0.6]},
                            {"label": "average", "values": [0.3, 0.6]},
                            {"label": "something", "values": [0.3, 0.6]},
                            {"label": "some string", "values": ["absc", "gar"]},
                        ],
                        "graph": {
                            "data": fig_dict["data"],
                            "layout": fig_dict["layout"],
                        },
                        "details": {
                            "parts": [
                                {"title": "First Tab", "id": "first_tab"},
                                {"title": "Second Tab", "id": "second_tab"},
                            ],
                            "insights": [],
                        },
                    },
                    additionalGraphs=[
                        AdditionalGraphInfo(
                            "first_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                        AdditionalGraphInfo(
                            "second_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                    ],
                ),
                BaseWidgetInfo(
                    type="rich_data",
                    title="",
                    size=2,
                    params={
                        "header": "Without Graph",
                        "description": "optional description",
                        "metricsValuesHeaders": ["reference", "current"],
                        "metrics": [
                            {"label": "f1", "values": [0.3, 0.6]},
                            {"label": "average", "values": [0.3, 0.6]},
                            {"label": "something", "values": [0.3, 0.6]},
                            {"label": "some string", "values": ["absc", "gar"]},
                        ],
                        "details": {
                            "parts": [
                                {"title": "First Tab", "id": "first_tab"},
                                {"title": "Second Tab", "id": "second_tab"},
                            ],
                            "insights": [],
                        },
                    },
                    additionalGraphs=[
                        AdditionalGraphInfo(
                            "first_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                        AdditionalGraphInfo(
                            "second_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                    ],
                ),
                BaseWidgetInfo(
                    type="rich_data",
                    title="",
                    size=2,
                    params={
                        "header": "Without details",
                        "description": "optional description",
                        "metricsValuesHeaders": ["reference", "current"],
                        "metrics": [
                            {"label": "f1", "values": [0.3, 0.6]},
                            {"label": "average", "values": [0.3, 0.6]},
                            {"label": "something", "values": [0.3, 0.6]},
                            {"label": "some string", "values": ["absc", "gar"]},
                        ],
                        "graph": {
                            "data": fig_dict["data"],
                            "layout": fig_dict["layout"],
                        },
                    },
                ),
                BaseWidgetInfo(
                    type="rich_data",
                    title="",
                    size=2,
                    params={
                        "header": "Without graph and details",
                        "description": "optional description",
                        "metricsValuesHeaders": ["reference", "current"],
                        "metrics": [
                            {"label": "f1", "values": [0.3, 0.6]},
                            {"label": "average", "values": [0.3, 0.6]},
                            {"label": "something", "values": [0.3, 0.6]},
                            {"label": "some string", "values": ["absc", "gar"]},
                        ],
                    },
                ),
                BaseWidgetInfo(
                    type="rich_data",
                    title="",
                    size=2,
                    params={
                        "header": "Header",
                        "description": "optional description",
                        "metricsValuesHeaders": ["reference", "current"],
                        "metrics": [
                            {"label": "f1", "values": [0.3, 0.6]},
                            {"label": "average", "values": [0.3, 0.6]},
                            {"label": "something", "values": [0.3, 0.6]},
                            {"label": "some string", "values": ["absc", "gar"]},
                        ],
                        "graph": {
                            "data": fig_dict["data"],
                            "layout": fig_dict["layout"],
                        },
                        "details": {
                            "parts": [
                                {"title": "First Tab", "id": "first_tab"},
                                {"title": "Second Tab", "id": "second_tab"},
                            ],
                            "insights": [],
                        },
                    },
                    additionalGraphs=[
                        AdditionalGraphInfo(
                            "first_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                        AdditionalGraphInfo(
                            "second_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                    ],
                ),
                BaseWidgetInfo(
                    type="rich_data",
                    title="",
                    size=2,
                    params={
                        "header": "More values",
                        "description": "optional description",
                        "metricsValuesHeaders": ["first", "second", "third"],
                        "metrics": [
                            {"label": "f1", "values": [0.3, 0.6, 0.9]},
                            {"label": "average", "values": [0.3, 0.6, 0.7]},
                            {"label": "something", "values": [0.3, 0.6, 0.9]},
                            {"label": "some string", "values": ["absc", "gar", "par"]},
                        ],
                        "graph": {
                            "data": fig_dict["data"],
                            "layout": fig_dict["layout"],
                        },
                        "details": {
                            "parts": [
                                {"title": "First Tab", "id": "first_tab"},
                                {"title": "Second Tab", "id": "second_tab"},
                            ],
                        },
                    },
                    additionalGraphs=[
                        AdditionalGraphInfo(
                            "first_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                        AdditionalGraphInfo(
                            "second_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                    ],
                ),
                BaseWidgetInfo(
                    type="rich_data",
                    title="",
                    size=2,
                    params={
                        "header": "With insight",
                        "description": "optional description",
                        "metricsValuesHeaders": ["reference", "current"],
                        "metrics": [
                            {"label": "f1", "values": [0.3, 0.6]},
                            {"label": "average", "values": [0.3, 0.6]},
                            {"label": "something", "values": [0.3, 0.6]},
                            {"label": "some string", "values": ["absc", "gar"]},
                        ],
                        "graph": {
                            "data": fig_dict["data"],
                            "layout": fig_dict["layout"],
                        },
                        "details": {
                            "parts": [
                                {"title": "First Tab", "id": "first_tab"},
                                {"title": "Second Tab", "id": "second_tab"},
                            ],
                            "insights": [
                                Insight(
                                    "Some notification",
                                    "warning",
                                    "additional text for insight",
                                )
                            ],
                        },
                    },
                    additionalGraphs=[
                        AdditionalGraphInfo(
                            "first_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                        AdditionalGraphInfo(
                            "second_tab",
                            {
                                "data": fig_dict["data"],
                                "layout": fig_dict["layout"],
                            },
                        ),
                    ],
                ),
            ],
            pageSize=2,
        )



import dataclasses
import json
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
from plotly import graph_objs as go
from plotly.subplots import make_subplots
from scipy.stats import probplot

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class RegressionErrorNormalityResults:
    current_error: pd.Series
    reference_error: Optional[pd.Series]


class RegressionErrorNormality(Metric[RegressionErrorNormalityResults]):
    def calculate(self, data: InputData) -> RegressionErrorNormalityResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        curr_df = data.current_data
        ref_df = data.reference_data
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        if not isinstance(prediction_name, str):
            raise ValueError("Expect one column for prediction. List of columns was provided.")
        curr_df = self._make_df_for_plot(curr_df, target_name, prediction_name, None)
        current_error = curr_df[prediction_name] - curr_df[target_name]
        reference_error = None
        if ref_df is not None:
            ref_df = self._make_df_for_plot(ref_df, target_name, prediction_name, None)
            reference_error = ref_df[prediction_name] - ref_df[target_name]
        return RegressionErrorNormalityResults(current_error=current_error, reference_error=reference_error)

    def _make_df_for_plot(self, df, target_name: str, prediction_name: str, datetime_column_name: Optional[str]):
        result = df.replace([np.inf, -np.inf], np.nan)
        if datetime_column_name is not None:
            result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name, datetime_column_name])
            return result.sort_values(datetime_column_name)
        result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name])
        return result.sort_index()


@default_renderer(wrap_type=RegressionErrorNormality)
class RegressionErrorNormalityRenderer(MetricRenderer):
    def render_json(self, obj: RegressionErrorNormality) -> dict:
        return {}

    def render_html(self, obj: RegressionErrorNormality) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        current_error = result.current_error
        reference_error = result.reference_error
        color_options = self.color_options
        cols = 1
        subplot_titles: Union[list, str] = ""

        if reference_error is not None:
            cols = 2
            subplot_titles = ["current", "reference"]

        fig = make_subplots(rows=1, cols=cols, shared_yaxes=False, subplot_titles=subplot_titles)
        curr_qq_lines = probplot(current_error, dist="norm", plot=None)
        curr_theoretical_q_x = np.linspace(curr_qq_lines[0][0][0], curr_qq_lines[0][0][-1], 100)
        sample_quantile_trace = go.Scatter(
            x=curr_qq_lines[0][0],
            y=curr_qq_lines[0][1],
            mode="markers",
            name="Dataset Quantiles",
            legendgroup="Dataset Quantiles",
            marker=dict(size=6, color=color_options.primary_color),
        )

        theoretical_quantile_trace = go.Scatter(
            x=curr_theoretical_q_x,
            y=curr_qq_lines[1][0] * curr_theoretical_q_x + curr_qq_lines[1][1],
            mode="lines",
            name="Theoretical Quantiles",
            legendgroup="Theoretical Quantiles",
            marker=dict(size=6, color=color_options.secondary_color),
        )
        fig.add_trace(sample_quantile_trace, 1, 1)
        fig.add_trace(theoretical_quantile_trace, 1, 1)
        fig.update_xaxes(title_text="Theoretical Quantiles", row=1, col=1)
        if reference_error is not None:
            ref_qq_lines = probplot(reference_error, dist="norm", plot=None)
            ref_theoretical_q_x = np.linspace(ref_qq_lines[0][0][0], ref_qq_lines[0][0][-1], 100)
            sample_quantile_trace = go.Scatter(
                x=ref_qq_lines[0][0],
                y=ref_qq_lines[0][1],
                mode="markers",
                name="Dataset Quantiles",
                legendgroup="Dataset Quantiles",
                showlegend=False,
                marker=dict(size=6, color=color_options.primary_color),
            )

            theoretical_quantile_trace = go.Scatter(
                x=ref_theoretical_q_x,
                y=ref_qq_lines[1][0] * ref_theoretical_q_x + ref_qq_lines[1][1],
                mode="lines",
                name="Theoretical Quantiles",
                legendgroup="Theoretical Quantiles",
                showlegend=False,
                marker=dict(size=6, color=color_options.secondary_color),
            )
            fig.add_trace(sample_quantile_trace, 1, 2)
            fig.add_trace(theoretical_quantile_trace, 1, 2)
            fig.update_xaxes(title_text="Theoretical Quantiles", row=1, col=2)
        fig.update_layout(yaxis_title="Dataset Quantiles")
        fig = json.loads(fig.to_json())

        return [
            header_text(label="Error Normality"),
            BaseWidgetInfo(
                title="",
                size=2,
                type="big_graph",
                params={"data": fig["data"], "layout": fig["layout"]},
            ),
        ]



import dataclasses
from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.classification_performance import get_prediction_data
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import get_class_separation_plot_data
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import widget_tabs
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ClassificationClassSeparationPlotResults:
    target_name: str
    current_plot: Optional[pd.DataFrame] = None
    reference_plot: Optional[pd.DataFrame] = None


class ClassificationClassSeparationPlot(Metric[ClassificationClassSeparationPlotResults]):
    def calculate(self, data: InputData) -> ClassificationClassSeparationPlotResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        curr_predictions = get_prediction_data(data.current_data, dataset_columns, data.column_mapping.pos_label)
        if curr_predictions.prediction_probas is None:
            raise ValueError(
                "ClassificationClassSeparationPlot can be calculated only on binary probabilistic predictions"
            )
        current_plot = curr_predictions.prediction_probas.copy()
        current_plot[target_name] = data.current_data[target_name]
        reference_plot = None
        if data.reference_data is not None:
            ref_predictions = get_prediction_data(data.reference_data, dataset_columns, data.column_mapping.pos_label)
            if ref_predictions.prediction_probas is None:
                raise ValueError(
                    "ClassificationClassSeparationPlot can be calculated only on binary probabilistic predictions"
                )
            reference_plot = ref_predictions.prediction_probas.copy()
            reference_plot[target_name] = data.reference_data[target_name]
        return ClassificationClassSeparationPlotResults(
            current_plot=current_plot,
            reference_plot=reference_plot,
            target_name=target_name,
        )


@default_renderer(wrap_type=ClassificationClassSeparationPlot)
class ClassificationClassSeparationPlotRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationClassSeparationPlot) -> dict:
        return {}

    def render_html(self, obj: ClassificationClassSeparationPlot) -> List[BaseWidgetInfo]:
        current_plot = obj.get_result().current_plot
        reference_plot = obj.get_result().reference_plot
        target_name = obj.get_result().target_name
        if current_plot is None:
            return []
        current_plot.replace([np.inf, -np.inf], np.nan, inplace=True)
        if reference_plot is not None:
            reference_plot.replace([np.inf, -np.inf], np.nan, inplace=True)
        tab_data = get_class_separation_plot_data(
            current_plot, reference_plot, target_name, color_options=self.color_options
        )
        tabs = [TabData(name, widget) for name, widget in tab_data]
        return [header_text(label="Class Separation Quality"), widget_tabs(title="", tabs=tabs)]



#!/usr/bin/env python
# coding: utf-8

from dataclasses import dataclass
from typing import List

from evidently.model.widget import BaseWidgetInfo


@dataclass
class DashboardInfo:
    name: str
    widgets: List[BaseWidgetInfo]



import abc
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type

import pandas

from evidently.analyzers.base_analyzer import Analyzer
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.pipeline.pipeline import Pipeline
from evidently.pipeline.stage import PipelineStage


class ModelMonitoringMetric:
    def __init__(self, name: str, labels: List[str] = None):
        self.name = name
        self.labels = labels if labels is not None else []

    def create(self, value: float, labels: Optional[Dict[str, str]] = None):
        _labels = labels if labels is not None else {}
        if len(_labels) == 0 and len(self.labels) == 0:
            return self, value, None
        if set(_labels.keys()) != set(self.labels):
            raise ValueError(
                f"Trying to create metric with"
                f" incorrect labels got {set(_labels.keys())} expected {set(self.labels)}"
            )
        return self, value, _labels


MetricsType = Tuple[ModelMonitoringMetric, float, Optional[Dict[str, str]]]


class ModelMonitor(PipelineStage):
    def calculate(
        self,
        reference_data: pandas.DataFrame,
        current_data: pandas.DataFrame,
        column_mapping: ColumnMapping,
        analyzers_results: Dict[Type[Analyzer], Any],
    ):
        pass

    @abc.abstractmethod
    def monitor_id(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def analyzers(self) -> List[Type[Analyzer]]:
        raise NotImplementedError()

    @abc.abstractmethod
    def metrics(self, analyzer_results):
        raise NotImplementedError()


class ModelMonitoring(Pipeline):
    def __init__(
        self, monitors: Sequence[ModelMonitor], options: Optional[list] = None
    ):
        if options is None:
            options = []

        super().__init__(monitors, options)
        self.monitors = list(monitors)

    def get_analyzers(self):
        return list({analyzer for tab in self.monitors for analyzer in tab.analyzers()})

    def metrics(self) -> Generator[MetricsType, None, None]:
        for monitor in self.monitors:
            for metric in monitor.metrics(self.analyzers_results):
                yield metric



import dataclasses
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
import plotly.figure_factory as ff
import sklearn
from plotly import graph_objs as go
from plotly.subplots import make_subplots

from evidently.base_metric import InputData
from evidently.metrics.classification_performance.base_classification_metric import ThresholdClassificationMetric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import WidgetSize
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ClassificationQualityByClassResult:
    columns: DatasetColumns
    current_metrics: dict
    current_roc_aucs: Optional[list]
    reference_metrics: Optional[dict]
    reference_roc_aucs: Optional[dict]


class ClassificationQualityByClass(ThresholdClassificationMetric[ClassificationQualityByClassResult]):
    def __init__(
        self,
        probas_threshold: Optional[float] = None,
        k: Optional[Union[float, int]] = None,
    ):
        super().__init__(probas_threshold, k)

    def calculate(self, data: InputData) -> ClassificationQualityByClassResult:
        columns = process_columns(data.current_data, data.column_mapping)
        target, prediction = self.get_target_prediction_data(
            data.current_data,
            column_mapping=data.column_mapping,
        )
        metrics_matrix = sklearn.metrics.classification_report(
            target,
            prediction.predictions,
            output_dict=True,
        )

        current_roc_aucs = None
        if prediction.prediction_probas is not None:
            binaraized_target = (target.values.reshape(-1, 1) == list(prediction.prediction_probas.columns)).astype(int)
            current_roc_aucs = sklearn.metrics.roc_auc_score(
                binaraized_target, prediction.prediction_probas, average=None
            ).tolist()
        ref_metrics = None
        reference_roc_aucs = None

        if data.reference_data is not None:
            ref_target, ref_prediction = self.get_target_prediction_data(
                data.reference_data,
                column_mapping=data.column_mapping,
            )
            ref_metrics = sklearn.metrics.classification_report(
                ref_target,
                ref_prediction.predictions,
                output_dict=True,
            )
            if ref_prediction.prediction_probas is not None:
                binaraized_target = (
                    ref_target.values.reshape(-1, 1) == list(ref_prediction.prediction_probas.columns)
                ).astype(int)
                reference_roc_aucs = sklearn.metrics.roc_auc_score(
                    binaraized_target, ref_prediction.prediction_probas, average=None
                ).tolist()
        return ClassificationQualityByClassResult(
            columns=columns,
            current_metrics=metrics_matrix,
            current_roc_aucs=current_roc_aucs,
            reference_metrics=ref_metrics,
            reference_roc_aucs=reference_roc_aucs,
        )


@default_renderer(wrap_type=ClassificationQualityByClass)
class ClassificationQualityByClassRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationQualityByClass) -> dict:
        return dataclasses.asdict(obj.get_result())

    def render_html(self, obj: ClassificationQualityByClass) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        columns = metric_result.columns
        current_metrics = metric_result.current_metrics
        current_roc_aucs = metric_result.current_roc_aucs
        reference_metrics = metric_result.reference_metrics
        reference_roc_aucs = metric_result.reference_roc_aucs

        metrics_frame = pd.DataFrame(current_metrics)
        z = metrics_frame.iloc[:-1, :-3].values
        x = columns.target_names if columns.target_names else metrics_frame.columns.tolist()[:-3]
        y = ["precision", "recall", "f1-score"]
        if current_roc_aucs is not None and len(current_roc_aucs) > 2:
            z = np.append(z, [current_roc_aucs], axis=0)
            y.append("roc-auc")

        # change each element of z to type string for annotations
        z_text = [[str(round(y, 3)) for y in x] for x in z]

        if reference_metrics is not None:
            cols = 2
            subplot_titles = ["current", "reference"]
        else:
            cols = 1
            subplot_titles = [""]
        fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
        trace = go.Heatmap(
            z=z,
            x=x,
            y=y,
            text=z_text,
            texttemplate="%{text}",
            coloraxis="coloraxis",
        )
        fig.add_trace(trace, 1, 1)

        if reference_metrics is not None:
            ref_metrics_frame = pd.DataFrame(reference_metrics)
            z = ref_metrics_frame.iloc[:-1, :-3].values
            x = columns.target_names if columns.target_names else metrics_frame.columns.tolist()[:-3]
            y = ["precision", "recall", "f1-score"]

            if current_roc_aucs is not None and len(current_roc_aucs) > 2:
                z = np.append(z, [reference_roc_aucs], axis=0)
                y.append("roc-auc")

            z_text = [[str(round(y, 3)) for y in x] for x in z]
            trace = go.Heatmap(
                z=z,
                x=x,
                y=y,
                text=z_text,
                texttemplate="%{text}",
                coloraxis="coloraxis",
            )
            fig.add_trace(trace, 1, 2)
        fig.update_layout(coloraxis={"colorscale": "RdBu_r"})

        return [header_text(label="Quality Metrics by Class"), plotly_figure(figure=fig, title="")]


def _plot_metrics(
    columns: DatasetColumns,
    metrics_matrix: dict,
    title: str,
    size: WidgetSize,
):
    metrics_frame = pd.DataFrame(metrics_matrix)

    z = metrics_frame.iloc[:-1, :-3].values

    x = columns.target_names if columns.target_names else metrics_frame.columns.tolist()[:-3]

    y = ["precision", "recall", "f1-score"]

    # change each element of z to type string for annotations
    z_text = [[str(round(y, 3)) for y in x] for x in z]

    # set up figure
    fig = ff.create_annotated_heatmap(z, x=x, y=y, annotation_text=z_text, colorscale="bluered", showscale=True)
    fig.update_layout(xaxis_title="Class", yaxis_title="Metric")

    return plotly_figure(
        title=title,
        figure=fig,
        size=size,
    )



import base64
import dataclasses
import json
import os
import shutil
from dataclasses import asdict
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional

import evidently
from evidently.model.dashboard import DashboardInfo
from evidently.utils import NumpyEncoder

STATIC_PATH = os.path.join(evidently.__path__[0], "nbextension", "static")


class SaveMode(Enum):
    SINGLE_FILE = "singlefile"
    FOLDER = "folder"
    SYMLINK_FOLDER = "symlink_folder"


SaveModeMap = {v.value: v for v in SaveMode}


@dataclasses.dataclass()
class TemplateParams:
    dashboard_id: str
    dashboard_info: DashboardInfo
    additional_graphs: Dict
    embed_font: bool = True
    embed_lib: bool = True
    embed_data: bool = True
    font_file: Optional[str] = None
    include_js_files: List[str] = dataclasses.field(default_factory=list)


def save_lib_files(filename: str, mode: SaveMode):
    if mode == SaveMode.SINGLE_FILE:
        return None, None
    parent_dir = os.path.dirname(filename)
    if not os.path.exists(os.path.join(parent_dir, "js")):
        os.makedirs(os.path.join(parent_dir, "js"), exist_ok=True)
    font_file = os.path.join(parent_dir, "js", "material-ui-icons.woff2")
    lib_file = os.path.join(parent_dir, "js", f"evidently.{evidently.__version__}.js")

    if mode == SaveMode.SYMLINK_FOLDER:
        if os.path.exists(font_file):
            os.remove(font_file)
        os.symlink(os.path.join(STATIC_PATH, "material-ui-icons.woff2"), font_file)
        if os.path.exists(lib_file):
            os.remove(lib_file)
        os.symlink(os.path.join(STATIC_PATH, "index.js"), lib_file)
    else:
        shutil.copy(os.path.join(STATIC_PATH, "material-ui-icons.woff2"), font_file)
        shutil.copy(os.path.join(STATIC_PATH, "index.js"), lib_file)
    return font_file, lib_file


def save_data_file(filename: str, mode: SaveMode, dashboard_id, dashboard_info: DashboardInfo, additional_graphs: Dict):
    if mode == SaveMode.SINGLE_FILE:
        return None
    parent_dir = os.path.dirname(filename)
    if parent_dir and not os.path.exists(parent_dir):
        os.makedirs(parent_dir, exist_ok=True)
    base_name = os.path.basename(filename)
    data_file = os.path.join(parent_dir, "js", f"{base_name}.data.js")
    with open(data_file, "w", encoding="utf-8") as out_file:
        out_file.write(
            f"""
    var {dashboard_id} = {dashboard_info_to_json(dashboard_info)};
    var additional_graphs_{dashboard_id} = {json.dumps(additional_graphs, cls=NumpyEncoder)};"""
        )
    return data_file


def dashboard_info_to_json(dashboard_info: DashboardInfo):
    asdict_result = asdict(dashboard_info)
    for widget in asdict_result["widgets"]:
        widget.pop("additionalGraphs", None)
    return json.dumps(asdict_result, cls=NumpyEncoder)


def inline_template(params: TemplateParams):
    return f"""
<link rel="stylesheet" href="https://fonts.googleapis.com/icon?family=Material+Icons"/>
<style>
.reset-this-parent {{
  all: initial;
}}
.reset-this-parent h5 {{
  all: initial;
  font: initial;
}}

svg {{
  height: intrinsic !important;
}}
</style>
<script>
    var {params.dashboard_id} = {dashboard_info_to_json(params.dashboard_info)};
    var additional_graphs_{params.dashboard_id} = {json.dumps(params.additional_graphs, cls=NumpyEncoder)};
</script>
<script>
function domReady(fn) {{
  // If we're early to the party
  document.addEventListener("DOMContentLoaded", fn);
  // If late; I mean on time.
  if (document.readyState === "interactive" || document.readyState === "complete" ) {{
    fn();
  }}
}}

domReady(function () {{
    requirejs(["evidently"], function(ev) {{
        drawDashboard({params.dashboard_id},
        new Map(Object.entries(additional_graphs_{params.dashboard_id})),
        "root_{params.dashboard_id}");
    }},
    function(err) {{
        $("#root_{params.dashboard_id}").innerHTML = "Failed to load";
    }})
}});
</script>
<div class="reset-this-parent" id="root_{params.dashboard_id}">Loading...</div>

"""


def file_html_template(params: TemplateParams):
    lib_block = f"""<script>{__load_js()}</script>""" if params.embed_lib else "<!-- no embedded lib -->"
    data_block = (
        f"""<script>
    var {params.dashboard_id} = {dashboard_info_to_json(params.dashboard_info)};
    var additional_graphs_{params.dashboard_id} = {json.dumps(params.additional_graphs, cls=NumpyEncoder)};
</script>"""
        if params.embed_data
        else "<!-- no embedded data -->"
    )
    js_files_block = "\n".join([f'<script src="{file}"></script>' for file in params.include_js_files])
    return f"""
<html>
<head>
<meta charset="utf-8">
<style>
/* fallback */
@font-face {{
  font-family: 'Material Icons';
  font-style: normal;
  font-weight: 400;
  src: {f"url(data:font/ttf;base64,{__load_font()}) format('woff2');" if params.embed_font else
    f"url({params.font_file});"}
}}

.material-icons {{
  font-family: 'Material Icons';
  font-weight: normal;
  font-style: normal;
  font-size: 24px;
  line-height: 1;
  letter-spacing: normal;
  text-transform: none;
  display: inline-block;
  white-space: nowrap;
  word-wrap: normal;
  direction: ltr;
  text-rendering: optimizeLegibility;
  -webkit-font-smoothing: antialiased;
}}
</style>
{data_block}
</head>
<body>
<div id="root_{params.dashboard_id}">Loading...</div>
{lib_block}
{js_files_block}
<script>
window.drawDashboard({params.dashboard_id},
    new Map(Object.entries(additional_graphs_{params.dashboard_id})),
    "root_{params.dashboard_id}"
);
</script>
</body>
"""


def __load_js():
    return open(os.path.join(STATIC_PATH, "index.js"), encoding="utf-8").read()


def __load_font():
    return base64.b64encode(open(os.path.join(STATIC_PATH, "material-ui-icons.woff2"), "rb").read()).decode()



import dataclasses
from typing import List
from typing import Optional
from typing import Union

from evidently.base_metric import InputData
from evidently.calculations.classification_performance import DatasetClassificationQuality
from evidently.calculations.classification_performance import calculate_metrics
from evidently.metrics.classification_performance.base_classification_metric import ThresholdClassificationMetric
from evidently.metrics.classification_performance.confusion_matrix_metric import ClassificationConfusionMatrix
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ClassificationQualityMetricResult:
    current: DatasetClassificationQuality
    reference: Optional[DatasetClassificationQuality]
    target_name: str


class ClassificationQualityMetric(ThresholdClassificationMetric[ClassificationQualityMetricResult]):
    confusion_matrix_metric: ClassificationConfusionMatrix

    def __init__(
        self,
        probas_threshold: Optional[float] = None,
        k: Optional[Union[float, int]] = None,
    ):
        super().__init__(probas_threshold=probas_threshold, k=k)
        self.confusion_matrix_metric = ClassificationConfusionMatrix(probas_threshold=probas_threshold, k=k)

    def calculate(self, data: InputData) -> ClassificationQualityMetricResult:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        target, prediction = self.get_target_prediction_data(data.current_data, data.column_mapping)
        current = calculate_metrics(
            data.column_mapping,
            self.confusion_matrix_metric.get_result().current_matrix,
            target,
            prediction,
        )

        reference = None
        if data.reference_data is not None:
            ref_matrix = self.confusion_matrix_metric.get_result().reference_matrix
            if ref_matrix is None:
                raise ValueError(f"Dependency {self.confusion_matrix_metric.__class__} should have reference data")
            target, prediction = self.get_target_prediction_data(data.reference_data, data.column_mapping)
            reference = calculate_metrics(
                data.column_mapping,
                ref_matrix,
                target,
                prediction,
            )

        return ClassificationQualityMetricResult(current=current, reference=reference, target_name=target_name)


@default_renderer(wrap_type=ClassificationQualityMetric)
class ClassificationQualityMetricRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationQualityMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    def render_html(self, obj: ClassificationQualityMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        target_name = metric_result.target_name
        result = []
        counters = [
            CounterData("Accuracy", f"{round(metric_result.current.accuracy, 3)}"),
            CounterData("Precision", f"{round(metric_result.current.precision, 3)}"),
            CounterData("Recall", f"{round(metric_result.current.recall, 3)}"),
            CounterData("F1", f"{round(metric_result.current.f1, 3)}"),
        ]
        if metric_result.current.roc_auc is not None and metric_result.current.log_loss is not None:
            counters.extend(
                [
                    CounterData("ROC AUC", f"{round(metric_result.current.roc_auc, 3)}"),
                    CounterData("LogLoss", f"{round(metric_result.current.log_loss, 3)}"),
                ]
            )
        result.append(header_text(label=f"Classification Model Performance. Target: '{target_name}’"))
        result.append(counter(title="Current: Model Quality Metrics", counters=counters))

        if metric_result.reference is not None:
            counters = [
                CounterData("Accuracy", f"{round(metric_result.reference.accuracy, 3)}"),
                CounterData("Precision", f"{round(metric_result.reference.precision, 3)}"),
                CounterData("Recall", f"{round(metric_result.reference.recall, 3)}"),
                CounterData("F1", f"{round(metric_result.reference.f1, 3)}"),
            ]
            if metric_result.reference.roc_auc is not None and metric_result.reference.log_loss is not None:
                counters.extend(
                    [
                        CounterData("ROC AUC", f"{round(metric_result.reference.roc_auc, 3)}"),
                        CounterData("LogLoss", f"{round(metric_result.reference.log_loss, 3)}"),
                    ]
                )
            result.append(counter(title="Reference: Model Quality Metrics", counters=counters))
        return result



import datetime
import json

import numpy as np
import pandas as pd

from evidently.utils.types import ApproxValue

_TYPES_MAPPING = (
    (
        (
            np.int_,
            np.intc,
            np.intp,
            np.int8,
            np.int16,
            np.int32,
            np.int64,
            np.uint8,
            np.uint16,
            np.uint32,
            np.uint64,
        ),
        int,
    ),
    ((np.float_, np.float16, np.float32, np.float64), float),
    ((np.ndarray,), lambda obj: obj.tolist()),
    ((np.bool_), bool),
    ((pd.Timedelta,), str),
    (
        (np.void, type(pd.NaT)),
        lambda obj: None,
    ),  # should be before datetime as NaT is subclass of datetime.
    ((pd.Timestamp, datetime.datetime, datetime.date), lambda obj: obj.isoformat()),
    # map ApproxValue to json value
    ((ApproxValue,), lambda obj: obj.as_dict()),
)


class NumpyEncoder(json.JSONEncoder):
    """Numpy and Pandas data types to JSON types encoder"""

    def default(self, o):
        """JSON converter calls the method when it cannot convert an object to a Python type
        Convert the object to a Python type

        If we cannot convert the object, leave the default `JSONEncoder` behaviour - raise a TypeError exception.
        """
        for types_list, python_type in _TYPES_MAPPING:
            if isinstance(o, types_list):
                return python_type(o)

        return json.JSONEncoder.default(self, o)



from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence

from evidently import TaskType
from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_drift import ensure_prediction_column_is_string
from evidently.calculations.stattests import PossibleStatTestType
from evidently.metric_preset.metric_preset import MetricPreset
from evidently.metrics import ColumnCorrelationsMetric
from evidently.metrics import ColumnDriftMetric
from evidently.metrics import ColumnValuePlot
from evidently.metrics import TargetByFeaturesTable
from evidently.utils.data_drift_utils import resolve_stattest_threshold
from evidently.utils.data_operations import DatasetColumns


class TargetDriftPreset(MetricPreset):
    """Metric preset for Target Drift analysis.

    Contains metrics:
    - ColumnDriftMetric - for target and prediction if present in datasets.
    - ColumnValuePlot - if task is regression.
    - ColumnCorrelationsMetric - for target and prediction if present in datasets.
    - TargetByFeaturesTable
    """

    columns: Optional[List[str]]
    stattest: Optional[PossibleStatTestType]
    cat_stattest: Optional[PossibleStatTestType]
    num_stattest: Optional[PossibleStatTestType]
    text_stattest: Optional[PossibleStatTestType]
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]]
    stattest_threshold: Optional[float]
    cat_stattest_threshold: Optional[float]
    num_stattest_threshold: Optional[float]
    text_stattest_threshold: Optional[float]
    per_column_stattest_threshold: Optional[Dict[str, float]]

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        super().__init__()
        self.columns = columns
        self.stattest = stattest
        self.cat_stattest = cat_stattest
        self.num_stattest = num_stattest
        self.text_stattest = text_stattest
        self.per_column_stattest = per_column_stattest
        self.stattest_threshold = stattest_threshold
        self.cat_stattest_threshold = cat_stattest_threshold
        self.num_stattest_threshold = num_stattest_threshold
        self.text_stattest_threshold = text_stattest_threshold
        self.per_column_stattest_threshold = per_column_stattest_threshold

    def generate_metrics(self, data: InputData, columns: DatasetColumns) -> Sequence[Metric]:
        target = columns.utility_columns.target
        prediction = columns.utility_columns.prediction
        result: List[Metric] = []
        columns_by_target = []
        prob_columns: Optional[Sequence[str]] = None

        if target is not None:
            columns_by_target.append(target)

            stattest, threshold = resolve_stattest_threshold(
                target,
                "cat" if columns.task == TaskType.CLASSIFICATION_TASK else "num",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            result.append(
                ColumnDriftMetric(
                    column_name=target,
                    stattest=stattest,
                    stattest_threshold=threshold,
                )
            )

            if data.column_mapping.is_regression_task():
                result.append(ColumnValuePlot(column_name=target))

            result.append(ColumnCorrelationsMetric(column_name=target))

        if prediction is not None:
            if data.column_mapping.is_classification_task():
                prediction_column = ensure_prediction_column_is_string(
                    prediction_column=columns.utility_columns.prediction,
                    current_data=data.current_data,
                    reference_data=data.reference_data,
                )

                if prediction_column is not None:
                    # in case that a new prediction column was created
                    if not isinstance(prediction, str):
                        prob_columns = prediction

                    prediction = prediction_column

            if isinstance(prediction, str):
                columns_by_target.append(prediction)
                stattest, threshold = resolve_stattest_threshold(
                    prediction,
                    "cat" if columns.task == TaskType.CLASSIFICATION_TASK else "num",
                    self.stattest,
                    self.cat_stattest,
                    self.num_stattest,
                    self.text_stattest,
                    self.per_column_stattest,
                    self.stattest_threshold,
                    self.cat_stattest_threshold,
                    self.num_stattest_threshold,
                    self.text_stattest_threshold,
                    self.per_column_stattest_threshold,
                )
                result.append(
                    ColumnDriftMetric(
                        column_name=prediction,
                        stattest=stattest,
                        stattest_threshold=threshold,
                    )
                )

                if prob_columns is not None:
                    for prob_column in prob_columns:
                        stattest, threshold = resolve_stattest_threshold(
                            prob_column,
                            "num",
                            self.stattest,
                            self.cat_stattest,
                            self.num_stattest,
                            self.text_stattest,
                            self.per_column_stattest,
                            self.stattest_threshold,
                            self.cat_stattest_threshold,
                            self.num_stattest_threshold,
                            self.text_stattest_threshold,
                            self.per_column_stattest_threshold,
                        )
                        result.append(
                            ColumnDriftMetric(
                                column_name=prob_column,
                                stattest=stattest,
                                stattest_threshold=threshold,
                            )
                        )

                result.append(ColumnCorrelationsMetric(column_name=prediction))

        if columns_by_target:
            result.append(TargetByFeaturesTable())

        return result



import dataclasses
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_drift import ColumnDataDriftMetrics
from evidently.calculations.data_drift import get_drift_for_columns
from evidently.calculations.stattests import PossibleStatTestType
from evidently.model.widget import BaseWidgetInfo
from evidently.options import DataDriftOptions
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import ColumnDefinition
from evidently.renderers.html_widgets import ColumnType
from evidently.renderers.html_widgets import RichTableDataRow
from evidently.renderers.html_widgets import RowDetails
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.html_widgets import rich_table_data
from evidently.renderers.html_widgets import table_data
from evidently.renderers.render_utils import get_distribution_plot_figure
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns
from evidently.utils.visualizations import plot_scatter_for_data_drift


@dataclass
class DataDriftTableResults:
    number_of_columns: int
    number_of_drifted_columns: int
    share_of_drifted_columns: float
    dataset_drift: bool
    drift_by_columns: Dict[str, ColumnDataDriftMetrics]
    dataset_columns: DatasetColumns


class DataDriftTable(Metric[DataDriftTableResults]):
    columns: Optional[List[str]]
    options: DataDriftOptions

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        self.columns = columns
        self.options = DataDriftOptions(
            all_features_stattest=stattest,
            cat_features_stattest=cat_stattest,
            num_features_stattest=num_stattest,
            text_features_stattest=text_stattest,
            per_feature_stattest=per_column_stattest,
            all_features_threshold=stattest_threshold,
            cat_features_threshold=cat_stattest_threshold,
            num_features_threshold=num_stattest_threshold,
            text_features_threshold=text_stattest_threshold,
            per_feature_threshold=per_column_stattest_threshold,
        )

    def get_parameters(self) -> tuple:
        return None if self.columns is None else tuple(self.columns), self.options

    def calculate(self, data: InputData) -> DataDriftTableResults:
        if data.reference_data is None:
            raise ValueError("Reference dataset should be present")

        dataset_columns = process_columns(data.reference_data, data.column_mapping)
        result = get_drift_for_columns(
            current_data=data.current_data,
            reference_data=data.reference_data,
            data_drift_options=self.options,
            dataset_columns=dataset_columns,
            columns=self.columns,
        )
        return DataDriftTableResults(
            number_of_columns=result.number_of_columns,
            number_of_drifted_columns=result.number_of_drifted_columns,
            share_of_drifted_columns=result.share_of_drifted_columns,
            dataset_drift=result.dataset_drift,
            drift_by_columns=result.drift_by_columns,
            dataset_columns=result.dataset_columns,
        )


@default_renderer(wrap_type=DataDriftTable)
class DataDriftTableRenderer(MetricRenderer):
    def render_json(self, obj: DataDriftTable) -> dict:
        result = dataclasses.asdict(obj.get_result())

        # remove pandas dataset values and other useless for JSON fields
        result.pop("dataset_columns", None)

        for column_name, data in result["drift_by_columns"].items():
            data.pop("current_distribution", None)
            data.pop("reference_distribution", None)
            data.pop("current_small_distribution", None)
            data.pop("reference_small_distribution", None)
            data.pop("current_correlations", None)
            data.pop("reference_correlations", None)
            data.pop("current_scatter", None)
            data.pop("x_name", None)
            data.pop("plot_shape", None)

        return result

    def _generate_column_params(self, column_name: str, data: ColumnDataDriftMetrics) -> Optional[RichTableDataRow]:
        details = RowDetails()
        if data.column_type == "text":
            if (
                data.typical_examples_cur is not None
                and data.typical_examples_ref is not None
                and data.typical_words_cur is not None
                and data.typical_words_ref is not None
            ):
                current_table_words = table_data(
                    title="",
                    column_names=["", ""],
                    data=[[el, ""] for el in data.typical_words_cur],
                )
                details.with_part("current: characteristic words", info=current_table_words)
                reference_table_words = table_data(
                    title="",
                    column_names=["", ""],
                    data=[[el, ""] for el in data.typical_words_ref],
                )
                details.with_part("reference: characteristic words", info=reference_table_words)
                current_table_examples = table_data(
                    title="",
                    column_names=["", ""],
                    data=[[el, ""] for el in data.typical_examples_cur],
                )
                details.with_part("current: characteristic examples", info=current_table_examples)
                reference_table_examples = table_data(
                    title="",
                    column_names=["", ""],
                    data=[[el, ""] for el in data.typical_examples_ref],
                )
                details.with_part("reference: characteristic examples", info=reference_table_examples)

            data_drift = "Detected" if data.drift_detected else "Not Detected"

            # tabs = [
            #     TabData(title="Carent Dataset: characteristic WORDS", widget=current_table_words),
            #     TabData(title="Reference Dataset: characteristic WORDS", widget=reference_table_words),
            #     TabData(title="Carent Dataset: characteristic EXAMPLES", widget=current_table_examples),
            #     TabData(title="Reference Dataset: characteristic EXAMPLES", widget=reference_table_examples),
            # ]
            return RichTableDataRow(
                details=details,
                fields={
                    "column_name": column_name,
                    "column_type": data.column_type,
                    "stattest_name": data.stattest_name,
                    # "reference_distribution": {},
                    # "current_distribution": {},
                    "data_drift": data_drift,
                    "drift_score": round(data.drift_score, 6),
                },
            )

        else:
            if (
                data.current_small_distribution is None
                or data.reference_small_distribution is None
                or data.current_distribution is None
                or data.reference_distribution is None
            ):
                return None

            current_small_hist = data.current_small_distribution
            ref_small_hist = data.reference_small_distribution
            data_drift = "Detected" if data.drift_detected else "Not Detected"
            if (
                data.column_type == "num"
                and data.current_scatter is not None
                and data.x_name is not None
                and data.plot_shape is not None
            ):
                scatter_fig = plot_scatter_for_data_drift(
                    curr_y=data.current_scatter[data.column_name],
                    curr_x=data.current_scatter[data.x_name],
                    y0=data.plot_shape["y0"],
                    y1=data.plot_shape["y1"],
                    y_name=data.column_name,
                    x_name=data.x_name,
                    color_options=self.color_options,
                )
                scatter = plotly_figure(title="", figure=scatter_fig)
                details.with_part("DATA DRIFT", info=scatter)
            fig = get_distribution_plot_figure(
                current_distribution=data.current_distribution,
                reference_distribution=data.reference_distribution,
                color_options=self.color_options,
            )
            distribution = plotly_figure(title="", figure=fig)
            details.with_part("DATA DISTRIBUTION", info=distribution)
            return RichTableDataRow(
                details=details,
                fields={
                    "column_name": column_name,
                    "column_type": data.column_type,
                    "stattest_name": data.stattest_name,
                    "reference_distribution": {"x": list(ref_small_hist[1]), "y": list(ref_small_hist[0])},
                    "current_distribution": {"x": list(current_small_hist[1]), "y": list(current_small_hist[0])},
                    "data_drift": data_drift,
                    "drift_score": round(data.drift_score, 6),
                },
            )

    def render_html(self, obj: DataDriftTable) -> List[BaseWidgetInfo]:
        results = obj.get_result()
        color_options = self.color_options
        target_column = results.dataset_columns.utility_columns.target
        prediction_column = results.dataset_columns.utility_columns.prediction

        # set params data
        params_data = []

        # sort columns by drift score
        all_columns = sorted(
            results.drift_by_columns.keys(),
            key=lambda x: results.drift_by_columns[x].drift_score,
            reverse=True,
        )
        # move target and prediction to the top of the table
        columns = []
        if target_column in all_columns:
            columns.append(target_column)
            all_columns.remove(target_column)
        if isinstance(prediction_column, str) and prediction_column in all_columns:
            columns.append(prediction_column)
            all_columns.remove(prediction_column)

        columns = columns + all_columns

        for column_name in columns:
            column_params = self._generate_column_params(column_name, results.drift_by_columns[column_name])

            if column_params is not None:
                params_data.append(column_params)

        drift_percents = round(results.share_of_drifted_columns * 100, 3)

        return [
            header_text(label="Data Drift Summary"),
            rich_table_data(
                title=f"Drift is detected for {drift_percents}% of columns "
                f"({results.number_of_drifted_columns} out of {results.number_of_columns}).",
                columns=[
                    ColumnDefinition("Column", "column_name"),
                    ColumnDefinition("Type", "column_type"),
                    ColumnDefinition(
                        "Reference Distribution",
                        "reference_distribution",
                        ColumnType.HISTOGRAM,
                        options={"xField": "x", "yField": "y", "color": color_options.primary_color},
                    ),
                    ColumnDefinition(
                        "Current Distribution",
                        "current_distribution",
                        ColumnType.HISTOGRAM,
                        options={"xField": "x", "yField": "y", "color": color_options.primary_color},
                    ),
                    ColumnDefinition("Data Drift", "data_drift"),
                    ColumnDefinition("Stat Test", "stattest_name"),
                    ColumnDefinition("Drift Score", "drift_score"),
                ],
                data=params_data,
            ),
        ]



#!/usr/bin/env python
# coding: utf-8

from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class CounterWidget(Widget):
    def analyzers(self):
        return []

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        return BaseWidgetInfo(
            type="counter",
            title=self.title,
            size=1,
            params={
                "counters": [
                    {"value": "7 out of 12 features", "label": "Data Drift Detected"}
                ]
            },
        )
